import os, itertools
import numpy as np
from ROSCO_toolbox.ofTools.util.FileTools import save_yaml

def save_case_matrix_direct(case_list, dir_matrix):
    ### assumes all elements of the list are dict for that case that has the same keys!
    if not os.path.exists(dir_matrix):
            os.makedirs(dir_matrix)
    ofh = open(os.path.join(dir_matrix,'case_matrix.txt'),'w')
    case = case_list[0]
    for key in case.keys():
        k = key[0]
        ofh.write("%s  " % k)
    ofh.write("\n")
    for key in case.keys():
        k = key[1]
        ofh.write("%s  " % k)
    ofh.write("\n")
    for i in range(len(case_list)):
        case = case_list[i]
        for key in case.keys():
            ofh.write(str(case[key]))
            ofh.write("  ")
        ofh.write("\n")
    ofh.close()


def save_case_matrix(matrix_out, change_vars, dir_matrix):
    # save matrix file
    if type(change_vars[0]) is tuple:
        n_header_lines = len(change_vars[0])
    else:
        change_vars = [(var,) for var in change_vars]
        n_header_lines = 1

    n_cases = np.shape(matrix_out)[0]
    matrix_out = np.hstack((np.asarray([[i] for i in range(n_cases)]), matrix_out))

    change_vars = [('Case_ID',)+('',)*(n_header_lines-1)] + change_vars
    # col_len = [max([len(val) for val in matrix_out[:,j]] + [len(change_vars[j][0]), len(change_vars[j][1])]) for j in range(len(change_vars))]
    col_len = [max([len(str(val)) for val in matrix_out[:,j]] + [len(change_vars[j][header_i]) for header_i in range(n_header_lines)]) for j in range(len(change_vars))]

    text_out = []
    for header_i in range(n_header_lines):
        text_out.append(''.join([val.center(col+2) for val, col in zip([var[header_i] for var in change_vars], col_len)])+'\n')

    for row in matrix_out:
        row_str = ''
        for val, col in zip(row, col_len):
            if val is not str:
                val = str(val)
            row_str += val.center(col+2)
        row_str += '\n'
        text_out.append(row_str)

    if not os.path.exists(dir_matrix):
            os.makedirs(dir_matrix)
    ofh = open(os.path.join(dir_matrix,'case_matrix.txt'),'w')
    for row in text_out:
        ofh.write(row)
    ofh.close()

def save_case_matrix_yaml(matrix_out, change_vars, dir_matrix, case_names):

    matrix_out_yaml = {}
    for var in change_vars:
        matrix_out_yaml[var] = []
    matrix_out_yaml['Case_ID'] = []
    matrix_out_yaml['Case_Name'] = []

    for i, row in enumerate(matrix_out):
        matrix_out_yaml['Case_ID'].append(i)
        matrix_out_yaml['Case_Name'].append(case_names[i])
        for val, var in zip(row, change_vars):
            if type(val) is list:
                if len(val) == 1:
                    val = val[0]
            if type(val) in [np.float32, np.float64, np.single, np.double, np.longdouble]:
                val = float(val)
            elif type(val) in [np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64, np.intc, np.uintc, np.uint]:
                val = int(val)
            elif type(val) in [np.array, np.ndarray]:
                val = val.tolist()
            elif type(val) in [np.str_]:
                val = str(val)
            # elif len(val) > 0:
            #     val = val.tolist()
            matrix_out_yaml[var].append(val)

    if not os.path.exists(dir_matrix):
        os.makedirs(dir_matrix)

    save_yaml(dir_matrix, 'case_matrix.yaml', matrix_out_yaml)

def case_naming(n_cases, namebase=None):
    # case naming
    case_name = [('%d'%i).zfill(len('%d'%(n_cases-1))) for i in range(n_cases)]
    if namebase:
        case_name = [namebase+'_'+caseid for caseid in case_name]

    return case_name

def convert_str(val):
    def try_type(val, data_type):
        try:
            data_type(val)
            return True
        except:
            return False
#        return isinstance(val, data_type)  ### this doesn't work b/c of numpy data types; they're not instances of base types
    def try_list(val):
        try:
            val[0]
            return True
        except:
            return False

    if try_type(val, int) and int(val) == float(val):
        return int(val)
    elif try_type(val, float):
        return float(val)
    elif val=='True':
        return True
    elif val=='False':
        return False
    # elif type(val)!=str and try_list(val):
    #     return ", ".join(['{:}'.format(i) for i in val])
    else:
        return val

def CaseGen_General(case_inputs, dir_matrix='', namebase='', save_matrix=True):
    """ Cartesian product to enumerate over all combinations of set of variables that are changed together"""

    # put case dict into lists
    change_vars = sorted(case_inputs.keys())
    change_vals = [case_inputs[var]['vals'] for var in change_vars]
    change_group = [case_inputs[var]['group'] for var in change_vars]

    # find number of groups and length of groups
    group_set = list(set(change_group))
    group_len = [len(change_vals[change_group.index(i)]) for i in group_set]

    # case matrix, as indices
    group_idx = [range(n) for n in group_len]
    matrix_idx = list(itertools.product(*group_idx))

    # index of each group
    matrix_group_idx = [np.where([group_i == group_j for group_j in change_group])[0].tolist() for group_i in group_set]

    # build final matrix of variable values
    matrix_out = []
    for i, row in enumerate(matrix_idx):
        row_out = [None]*len(change_vars)
        for j, val in enumerate(row):
            for g in matrix_group_idx[j]:
                row_out[g] = change_vals[g][val]
        matrix_out.append(row_out)
    try:
        matrix_out = np.asarray(matrix_out, dtype=str)
    except:
        matrix_out = np.asarray(matrix_out)
    n_cases = np.shape(matrix_out)[0]

    # case naming
    case_name = case_naming(n_cases, namebase=namebase)
    
    # Save case matrix
    if save_matrix:
        if not dir_matrix:
            dir_matrix = os.getcwd()
        try:
            save_case_matrix(matrix_out, change_vars, dir_matrix)
            save_case_matrix_yaml(matrix_out, change_vars, dir_matrix, case_name)
        except: 
            save_case_matrix_yaml(matrix_out, change_vars, dir_matrix, case_name)

    case_list = []
    for i in range(n_cases):
        case_list_i = {}
        for j, var in enumerate(change_vars):
            case_list_i[var] = convert_str(matrix_out[i,j])
        case_list.append(case_list_i)


    return case_list, case_name


if __name__ == "__main__":

    case_inputs = {}
    case_inputs[("Fst","TMax")] = {'vals':[10.], 'group':0}
    case_inputs[("InflowWind","WindType")] = {'vals':[1], 'group':0}

    case_inputs[("InflowWind","HWindSpeed")] = {'vals':[8., 9., 10., 11., 12.], 'group':1}
    case_inputs[("ElastoDyn","RotSpeed")] = {'vals':[9.156, 10.296, 11.431, 11.89, 12.1], 'group':1}
    case_inputs[("ElastoDyn","BlPitch1")] = {'vals':[0., 0., 0., 0., 3.823], 'group':1}
    case_inputs[("ElastoDyn","BlPitch2")] = case_inputs[("ElastoDyn","BlPitch1")]
    case_inputs[("ElastoDyn","BlPitch3")] = case_inputs[("ElastoDyn","BlPitch1")]

    case_inputs[("ElastoDyn","GenDOF")] = {'vals':['True','False'], 'group':2}

    case_list, case_name = CaseGen_General(case_inputs, 'C:/Users/egaertne/WISDEM/AeroelasticSE/src/AeroelasticSE/', 'testing')



import os
import subprocess
class Turbsim_wrapper(object):
    def __init__(self):
        self.turbsim_exe = 'turbsim'
        #self.turbsim_input = './turbsim_default.in'
        self.turbsim_input = "/Users/pgraf/work/wese/turbsim_studies/pghacking/TurbSim.inp"
        self.run_dir = '.'
        self.debug_level = 2

    def execute(self):
        exec_string = [self.turbsim_exe, self.turbsim_input]
        olddir = os.getcwd()
        os.chdir(self.run_dir)

        if self.debug_level > 0:
            print ("EXECUTING TurbSim")
            print ("Executable: \t", self.turbsim_exe )
            print ("Run directory: \t", self.run_dir)
            print ("Input file: \t", self.turbsim_input)
            print ("Exec string: \t", exec_string)

        if self.debug_level > 1:
            subprocess.call(exec_string)
        else:
            FNULL = open(os.devnull, 'w')
            subprocess.call(exec_string, stdout=FNULL, stderr=subprocess.STDOUT)

        if self.debug_level > 0:
            print ("COMPLETE TurbSim")

        os.chdir(olddir)
if __name__=='__main__':
    wrapper = Turbsim_wrapper()
    wrapper.turbsim_exe = '/Users/jquick/TurbSim/bin/TurbSim_glin64'
    wrapper.execute()



# Copyright 2019 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
"""
Class FileProcessing used to write out controller 
    parameter files need to run ROSCO

Methods:
-----------
run_openfast
write_DISCON
read_DISCON
write_rotor_performance
load_from_txt
DISCON_dict
list_check
"""
import datetime
import os
import numpy as np
import subprocess
import ROSCO_toolbox

# Some useful constants
now = datetime.datetime.now()
pi = np.pi
rad2deg = np.rad2deg(1)
deg2rad = np.deg2rad(1)
rpm2RadSec = 2.0*(np.pi)/60.0
RadSec2rpm = 60/(2.0 * np.pi)

def write_DISCON(turbine, controller, param_file='DISCON.IN', txt_filename='Cp_Ct_Cq.txt', rosco_vt = {}):
    """
    Print the controller parameters to the DISCON.IN input file for the generic controller

    Parameters:
    -----------
    turbine: class
                Turbine class containing turbine operation information (ref speeds, etc...)
    controller: class
                Controller class containing controller operation information (gains, etc...)
    param_file: str, optional
        filename for parameter input file, should be DISCON.IN
    txt_filename: str, optional
                    filename of rotor performance file
    """

    # Get ROSCO var tree if not provided
    if not rosco_vt:
        rosco_vt = DISCON_dict(turbine, controller, txt_filename)

    print('Writing new controller parameter file parameter file: %s.' % param_file)
    # Should be obvious what's going on here...
    file = open(param_file,'w')
    file.write('! Controller parameter input file for the %s wind turbine\n' % turbine.TurbineName)
    file.write('!    - File written using ROSCO version {} controller tuning logic on {}\n'.format(ROSCO_toolbox.__version__, now.strftime('%m/%d/%y')))
    file.write('\n')
    file.write('!------- DEBUG ------------------------------------------------------------\n')
    file.write('{0:<12d}        ! LoggingLevel		- {{0: write no debug files, 1: write standard output .dbg-file, 2: write standard output .dbg-file and complete avrSWAP-array .dbg2-file}}\n'.format(int(rosco_vt['LoggingLevel'])))
    file.write('\n')
    file.write('!------- CONTROLLER FLAGS -------------------------------------------------\n')
    file.write('{0:<12d}        ! F_LPFType			- {{1: first-order low-pass filter, 2: second-order low-pass filter}}, [rad/s] (currently filters generator speed and pitch control signals\n'.format(int(rosco_vt['F_LPFType'])))
    file.write('{0:<12d}        ! F_NotchType		- Notch on the measured generator speed and/or tower fore-aft motion (for floating) {{0: disable, 1: generator speed, 2: tower-top fore-aft motion, 3: generator speed and tower-top fore-aft motion}}\n'.format(int(rosco_vt['F_NotchType'])))
    file.write('{0:<12d}        ! IPC_ControlMode	- Turn Individual Pitch Control (IPC) for fatigue load reductions (pitch contribution) {{0: off, 1: 1P reductions, 2: 1P+2P reductions}}\n'.format(int(rosco_vt['IPC_ControlMode'])))
    file.write('{0:<12d}        ! VS_ControlMode	- Generator torque control mode in above rated conditions {{0: constant torque, 1: constant power, 2: TSR tracking PI control with constant torque, 3: TSR tracking PI control with constant power}}\n'.format(int(rosco_vt['VS_ControlMode'])))
    file.write('{0:<12d}        ! PC_ControlMode    - Blade pitch control mode {{0: No pitch, fix to fine pitch, 1: active PI blade pitch control}}\n'.format(int(rosco_vt['PC_ControlMode'])))
    file.write('{0:<12d}        ! Y_ControlMode		- Yaw control mode {{0: no yaw control, 1: yaw rate control, 2: yaw-by-IPC}}\n'.format(int(rosco_vt['Y_ControlMode'])))
    file.write('{0:<12d}        ! SS_Mode           - Setpoint Smoother mode {{0: no setpoint smoothing, 1: introduce setpoint smoothing}}\n'.format(int(rosco_vt['SS_Mode'])))
    file.write('{0:<12d}        ! WE_Mode           - Wind speed estimator mode {{0: One-second low pass filtered hub height wind speed, 1: Immersion and Invariance Estimator, 2: Extended Kalman Filter}}\n'.format(int(rosco_vt['WE_Mode'])))
    file.write('{0:<12d}        ! PS_Mode           - Pitch saturation mode {{0: no pitch saturation, 1: implement pitch saturation}}\n'.format(int(rosco_vt['PS_Mode'])))
    file.write('{0:<12d}        ! SD_Mode           - Shutdown mode {{0: no shutdown procedure, 1: pitch to max pitch at shutdown}}\n'.format(int(rosco_vt['SD_Mode'])))
    file.write('{0:<12d}        ! Fl_Mode           - Floating specific feedback mode {{0: no nacelle velocity feedback, 1: feed back translational velocity, 2: feed back rotational veloicty}}\n'.format(int(rosco_vt['Fl_Mode'])))
    file.write('{0:<12d}        ! Flp_Mode          - Flap control mode {{0: no flap control, 1: steady state flap angle, 2: Proportional flap control, 2: Cyclic (1P) flap control}}\n'.format(int(rosco_vt['Flp_Mode'])))
    file.write('{0:<12d}        ! OL_Mode           - Open loop control mode {{0: no open loop control, 1: open loop control vs. time, 2: open loop control vs. wind speed}}\n'.format(int(rosco_vt['OL_Mode'])))
    file.write('\n')
    file.write('!------- FILTERS ----------------------------------------------------------\n') 
    file.write('{:<13.5f}       ! F_LPFCornerFreq	- Corner frequency (-3dB point) in the low-pass filters, [rad/s]\n'.format(rosco_vt['F_LPFCornerFreq'])) 
    file.write('{:<13.5f}       ! F_LPFDamping		- Damping coefficient {{used only when F_FilterType = 2}} [-]\n'.format(rosco_vt['F_LPFDamping']))
    file.write('{:<13.5f}       ! F_NotchCornerFreq	- Natural frequency of the notch filter, [rad/s]\n'.format(rosco_vt['F_NotchCornerFreq']))
    file.write('{} ! F_NotchBetaNumDen	- Two notch damping values (numerator and denominator, resp) - determines the width and depth of the notch, [-]\n'.format(''.join('{:<4.6f}  '.format(rosco_vt['F_NotchBetaNumDen'][i]) for i in range(len(rosco_vt['F_NotchBetaNumDen'])))))
    file.write('{:<13.5f}       ! F_SSCornerFreq    - Corner frequency (-3dB point) in the first order low pass filter for the setpoint smoother, [rad/s].\n'.format(rosco_vt['F_SSCornerFreq']))
    file.write('{:<13.5f}       ! F_WECornerFreq    - Corner frequency (-3dB point) in the first order low pass filter for the wind speed estimate [rad/s].\n'.format(rosco_vt['F_WECornerFreq']))
    file.write('{} ! F_FlCornerFreq    - Natural frequency and damping in the second order low pass filter of the tower-top fore-aft motion for floating feedback control [rad/s, -].\n'.format(''.join('{:<4.6f}  '.format(rosco_vt['F_FlCornerFreq'][i]) for i in range(len(rosco_vt['F_FlCornerFreq'])))))
    file.write('{:<13.5f}       ! F_FlHighPassFreq    - Natural frequency of first-order high-pass filter for nacelle fore-aft motion [rad/s].\n'.format(rosco_vt['F_FlHighPassFreq']))
    file.write('{} ! F_FlpCornerFreq   - Corner frequency and damping in the second order low pass filter of the blade root bending moment for flap control [rad/s, -].\n'.format(''.join('{:<4.6f}  '.format(rosco_vt['F_FlpCornerFreq'][i]) for i in range(len(rosco_vt['F_FlpCornerFreq'])))))
    
    file.write('\n')
    file.write('!------- BLADE PITCH CONTROL ----------------------------------------------\n')
    file.write('{:<11d}         ! PC_GS_n			- Amount of gain-scheduling table entries\n'.format(int(rosco_vt['PC_GS_n'])))
    file.write('{}              ! PC_GS_angles	    - Gain-schedule table: pitch angles [rad].\n'.format(''.join('{:<4.6f}  '.format(rosco_vt['PC_GS_angles'][i]) for i in range(len(rosco_vt['PC_GS_angles'])))))            
    file.write('{}              ! PC_GS_KP		- Gain-schedule table: pitch controller kp gains [s].\n'.format(''.join('{:<4.6f}  '.format(rosco_vt['PC_GS_KP'][i]) for i in range(len(rosco_vt['PC_GS_KP'])))))
    file.write('{}              ! PC_GS_KI		- Gain-schedule table: pitch controller ki gains [-].\n'.format(''.join('{:<4.6f}  '.format(rosco_vt['PC_GS_KI'][i]) for i in range(len(rosco_vt['PC_GS_KI'])))))
    file.write('{}              ! PC_GS_KD			- Gain-schedule table: pitch controller kd gains\n'.format(''.join('{:<4.6f}  '.format(rosco_vt['PC_GS_KD'][i]) for i in range(len(rosco_vt['PC_GS_KD'])))))
    file.write('{}              ! PC_GS_TF			- Gain-schedule table: pitch controller tf gains (derivative filter)\n'.format(''.join('{:<4.6f}  '.format(rosco_vt['PC_GS_TF'][i]) for i in range(len(rosco_vt['PC_GS_TF'])))))
    file.write('{:<014.5f}      ! PC_MaxPit			- Maximum physical pitch limit, [rad].\n'.format(rosco_vt['PC_MaxPit']))
    file.write('{:<014.5f}      ! PC_MinPit			- Minimum physical pitch limit, [rad].\n'.format(rosco_vt['PC_MinPit']))
    file.write('{:<014.5f}      ! PC_MaxRat			- Maximum pitch rate (in absolute value) in pitch controller, [rad/s].\n'.format(rosco_vt['PC_MaxRat']))
    file.write('{:<014.5f}      ! PC_MinRat			- Minimum pitch rate (in absolute value) in pitch controller, [rad/s].\n'.format(rosco_vt['PC_MinRat']))
    file.write('{:<014.5f}      ! PC_RefSpd			- Desired (reference) HSS speed for pitch controller, [rad/s].\n'.format(rosco_vt['PC_RefSpd']))
    file.write('{:<014.5f}      ! PC_FinePit		- Record 5: Below-rated pitch angle set-point, [rad]\n'.format(rosco_vt['PC_FinePit']))
    file.write('{:<014.5f}      ! PC_Switch			- Angle above lowest minimum pitch angle for switch, [rad]\n'.format(rosco_vt['PC_Switch']))
    file.write('\n')
    file.write('!------- INDIVIDUAL PITCH CONTROL -----------------------------------------\n')
    file.write('{:<13.1f}       ! IPC_IntSat		- Integrator saturation (maximum signal amplitude contribution to pitch from IPC), [rad]\n'.format(rosco_vt['IPC_IntSat'])) # Hardcode to 5 degrees
    file.write('{} ! IPC_KP			- Proportional gain for the individual pitch controller: first parameter for 1P reductions, second for 2P reductions, [-]\n'.format(''.join('{:<4.3e} '.format(rosco_vt['IPC_KP'][i]) for i in range(len(rosco_vt['IPC_KP'])))))
    file.write('{} ! IPC_KI			- Integral gain for the individual pitch controller: first parameter for 1P reductions, second for 2P reductions, [-]\n'.format(''.join('{:<4.3e} '.format(rosco_vt['IPC_KI'][i]) for i in range(len(rosco_vt['IPC_KI'])))))
    file.write('{} ! IPC_aziOffset		- Phase offset added to the azimuth angle for the individual pitch controller, [rad]. \n'.format(''.join('{:<4.6f}  '.format(rosco_vt['IPC_aziOffset'][i]) for i in range(len(rosco_vt['IPC_aziOffset'])))))
    file.write('{:<13.1f}       ! IPC_CornerFreqAct - Corner frequency of the first-order actuators model, to induce a phase lag in the IPC signal {{0: Disable}}, [rad/s]\n'.format(rosco_vt['IPC_CornerFreqAct']))
    file.write('\n')
    file.write('!------- VS TORQUE CONTROL ------------------------------------------------\n')
    file.write('{:<014.5f}      ! VS_GenEff			- Generator efficiency mechanical power -> electrical power, [should match the efficiency defined in the generator properties!], [%]\n'.format(rosco_vt['VS_GenEff']))
    file.write('{:<014.5f}      ! VS_ArSatTq		- Above rated generator torque PI control saturation, [Nm]\n'.format(rosco_vt['VS_ArSatTq']))
    file.write('{:<014.5f}      ! VS_MaxRat			- Maximum torque rate (in absolute value) in torque controller, [Nm/s].\n'.format(rosco_vt['VS_MaxRat']))
    file.write('{:<014.5f}      ! VS_MaxTq			- Maximum generator torque in Region 3 (HSS side), [Nm].\n'.format(rosco_vt['VS_MaxTq']))
    file.write('{:<014.5f}      ! VS_MinTq			- Minimum generator torque (HSS side), [Nm].\n'.format(rosco_vt['VS_MinTq']))
    file.write('{:<014.5f}      ! VS_MinOMSpd		- Minimum generator speed [rad/s]\n'.format(rosco_vt['VS_MinOMSpd']))
    file.write('{:<014.5f}      ! VS_Rgn2K			- Generator torque constant in Region 2 (HSS side), [Nm/(rad/s)^2]\n'.format(rosco_vt['VS_Rgn2K']))
    file.write('{:<014.5f}      ! VS_RtPwr			- Wind turbine rated power [W]\n'.format(rosco_vt['VS_RtPwr']))
    file.write('{:<014.5f}      ! VS_RtTq			- Rated torque, [Nm].\n'.format(rosco_vt['VS_RtTq']))
    file.write('{:<014.5f}      ! VS_RefSpd			- Rated generator speed [rad/s]\n'.format(rosco_vt['VS_RefSpd']))
    file.write('{:<11d}         ! VS_n				- Number of generator PI torque controller gains\n'.format(int(rosco_vt['VS_n'])))
    file.write('{:<014.5f}      ! VS_KP				- Proportional gain for generator PI torque controller [-]. (Only used in the transitional 2.5 region if VS_ControlMode =/ 2)\n'.format(rosco_vt['VS_KP']))
    file.write('{:<014.5f}      ! VS_KI				- Integral gain for generator PI torque controller [s]. (Only used in the transitional 2.5 region if VS_ControlMode =/ 2)\n'.format(rosco_vt['VS_KI']))
    file.write('{:<13.2f}       ! VS_TSRopt			- Power-maximizing region 2 tip-speed-ratio [rad].\n'.format(rosco_vt['VS_TSRopt']))
    file.write('\n')
    file.write('!------- SETPOINT SMOOTHER ---------------------------------------------\n')
    file.write('{:<13.5f}       ! SS_VSGain         - Variable speed torque controller setpoint smoother gain, [-].\n'.format(rosco_vt['SS_VSGain']))
    file.write('{:<13.5f}       ! SS_PCGain         - Collective pitch controller setpoint smoother gain, [-].\n'.format(rosco_vt['SS_PCGain']))
    file.write('\n')
    file.write('!------- WIND SPEED ESTIMATOR ---------------------------------------------\n')
    file.write('{:<13.3f}       ! WE_BladeRadius	- Blade length (distance from hub center to blade tip), [m]\n'.format(rosco_vt['WE_BladeRadius']))
    file.write('{:<11d}         ! WE_CP_n			- Amount of parameters in the Cp array\n'.format(int(rosco_vt['WE_CP_n'])))
    file.write('{}              ! WE_CP - Parameters that define the parameterized CP(lambda) function\n'.format(rosco_vt['WE_CP']))
    file.write('{:<13.1f}		! WE_Gamma			- Adaption gain of the wind speed estimator algorithm [m/rad]\n'.format(rosco_vt['WE_Gamma']))
    file.write('{:<13.1f}       ! WE_GearboxRatio	- Gearbox ratio [>=1],  [-]\n'.format(rosco_vt['WE_GearboxRatio']))
    file.write('{:<14.5f}       ! WE_Jtot			- Total drivetrain inertia, including blades, hub and casted generator inertia to LSS, [kg m^2]\n'.format(rosco_vt['WE_Jtot']))
    file.write('{:<13.3f}       ! WE_RhoAir			- Air density, [kg m^-3]\n'.format(rosco_vt['WE_RhoAir']))
    file.write(      '"{}"      ! PerfFileName      - File containing rotor performance tables (Cp,Ct,Cq) (absolute path or relative to this file)\n'.format(rosco_vt['PerfFileName']))
    file.write('{:<7d} {:<10d}  ! PerfTableSize     - Size of rotor performance tables, first number refers to number of blade pitch angles, second number referse to number of tip-speed ratios\n'.format(int(rosco_vt['PerfTableSize'][0]),int(rosco_vt['PerfTableSize'][1])))
    file.write('{:<11d}         ! WE_FOPoles_N      - Number of first-order system poles used in EKF\n'.format(int(rosco_vt['WE_FOPoles_N'])))
    file.write('{}              ! WE_FOPoles_v      - Wind speeds corresponding to first-order system poles [m/s]\n'.format(''.join('{:<4.4f} '.format(rosco_vt['WE_FOPoles_v'][i]) for i in range(len(rosco_vt['WE_FOPoles_v'])))))
    file.write('{}              ! WE_FOPoles        - First order system poles [1/s]\n'.format(''.join('{:<10.8f} '.format(rosco_vt['WE_FOPoles'][i]) for i in range(len(rosco_vt['WE_FOPoles'])))))
    file.write('\n')
    file.write('!------- YAW CONTROL ------------------------------------------------------\n')
    file.write('{:<13.5f}       ! Y_ErrThresh		- Yaw error threshold. Turbine begins to yaw when it passes this. [rad^2 s]\n'.format(rosco_vt['Y_ErrThresh']))
    file.write('{:<13.5f}       ! Y_IPC_IntSat		- Integrator saturation (maximum signal amplitude contribution to pitch from yaw-by-IPC), [rad]\n'.format(rosco_vt['Y_IPC_IntSat']))
    file.write('{:<11d}         ! Y_IPC_n			- Number of controller gains (yaw-by-IPC)\n'.format(int(rosco_vt['Y_IPC_n'])))
    file.write('{:<13.5f}       ! Y_IPC_KP			- Yaw-by-IPC proportional controller gain Kp\n'.format(rosco_vt['Y_IPC_KP']))
    file.write('{:<13.5f}       ! Y_IPC_KI			- Yaw-by-IPC integral controller gain Ki\n'.format(rosco_vt['Y_IPC_KI']))
    file.write('{:<13.5f}       ! Y_IPC_omegaLP		- Low-pass filter corner frequency for the Yaw-by-IPC controller to filtering the yaw alignment error, [rad/s].\n'.format(rosco_vt['Y_IPC_omegaLP']))
    file.write('{:<13.5f}       ! Y_IPC_zetaLP		- Low-pass filter damping factor for the Yaw-by-IPC controller to filtering the yaw alignment error, [-].\n'.format(rosco_vt['Y_IPC_zetaLP']))
    file.write('{:<13.5f}       ! Y_MErrSet			- Yaw alignment error, set point [rad]\n'.format(rosco_vt['Y_MErrSet']))
    file.write('{:<13.5f}       ! Y_omegaLPFast		- Corner frequency fast low pass filter, 1.0 [rad/s]\n'.format(rosco_vt['Y_omegaLPFast']))
    file.write('{:<13.5f}       ! Y_omegaLPSlow		- Corner frequency slow low pass filter, 1/60 [rad/s]\n'.format(rosco_vt['Y_omegaLPSlow']))
    file.write('{:<13.5f}       ! Y_Rate			- Yaw rate [rad/s]\n'.format(rosco_vt['Y_Rate']))
    file.write('\n')
    file.write('!------- TOWER FORE-AFT DAMPING -------------------------------------------\n')
    file.write('{:<11d}         ! FA_KI				- Integral gain for the fore-aft tower damper controller, -1 = off / >0 = on [rad s/m] - !NJA - Make this a flag\n'.format(int(rosco_vt['FA_KI'] )))
    file.write('{:<13.1f}       ! FA_HPFCornerFreq	- Corner frequency (-3dB point) in the high-pass filter on the fore-aft acceleration signal [rad/s]\n'.format(rosco_vt['FA_HPFCornerFreq'] ))
    file.write('{:<13.1f}       ! FA_IntSat			- Integrator saturation (maximum signal amplitude contribution to pitch from FA damper), [rad]\n'.format(rosco_vt['FA_IntSat'] ))
    file.write('\n')
    file.write('!------- MINIMUM PITCH SATURATION -------------------------------------------\n')
    file.write('{:<11d}         ! PS_BldPitchMin_N  - Number of values in minimum blade pitch lookup table (should equal number of values in PS_WindSpeeds and PS_BldPitchMin)\n'.format(int(rosco_vt['PS_BldPitchMin_N'])))
    file.write('{}              ! PS_WindSpeeds     - Wind speeds corresponding to minimum blade pitch angles [m/s]\n'.format(''.join('{:<4.4f} '.format(rosco_vt['PS_WindSpeeds'][i]) for i in range(len(rosco_vt['PS_WindSpeeds'])))))
    file.write('{}              ! PS_BldPitchMin    - Minimum blade pitch angles [rad]\n'.format(''.join('{:<10.8f} '.format(rosco_vt['PS_BldPitchMin'][i]) for i in range(len(rosco_vt['PS_BldPitchMin'])))))
    file.write('\n')
    file.write('!------- SHUTDOWN -----------------------------------------------------------\n')
    file.write('{:<014.5f}      ! SD_MaxPit         - Maximum blade pitch angle to initiate shutdown, [rad]\n'.format(rosco_vt['SD_MaxPit']))
    file.write('{:<014.5f}      ! SD_CornerFreq     - Cutoff Frequency for first order low-pass filter for blade pitch angle, [rad/s]\n'.format(rosco_vt['SD_CornerFreq']))
    file.write('\n')
    file.write('!------- Floating -----------------------------------------------------------\n')
    if rosco_vt['Fl_Mode'] == 2:
        floatstr = 'pitching'
    else:
        floatstr = 'velocity'
    file.write('{:<014.5f}      ! Fl_Kp             - Nacelle {} proportional feedback gain [s]\n'.format(rosco_vt['Fl_Kp'], floatstr))
    file.write('\n')
    file.write('!------- FLAP ACTUATION -----------------------------------------------------\n')
    file.write('{:<014.5f}      ! Flp_Angle         - Initial or steady state flap angle [rad]\n'.format(rosco_vt['Flp_Angle']))
    file.write('{:<014.8e}      ! Flp_Kp            - Blade root bending moment proportional gain for flap control [s]\n'.format(rosco_vt['Flp_Kp']))
    file.write('{:<014.8e}      ! Flp_Ki            - Flap displacement integral gain for flap control [-]\n'.format(rosco_vt['Flp_Ki']))
    file.write('{:<014.5f}      ! Flp_MaxPit        - Maximum (and minimum) flap pitch angle [rad]\n'.format(rosco_vt['Flp_MaxPit']))
    file.write('\n')
    file.write('!------- Open Loop Control -----------------------------------------------------\n')
    file.write('"{}"            ! OL_Filename       - Input file with open loop timeseries (absolute path or relative to this file)\n'.format(rosco_vt['OL_Filename']))
    file.write('{0:<12d}        ! Ind_Breakpoint    - The column in OL_Filename that contains the breakpoint (time if OL_Mode = 1)\n'.format(int(rosco_vt['Ind_Breakpoint'])))
    file.write('{0:<12d}        ! Ind_BldPitch      - The column in OL_Filename that contains the blade pitch input in rad\n'.format(int(rosco_vt['Ind_BldPitch'])))
    file.write('{0:<12d}        ! Ind_GenTq         - The column in OL_Filename that contains the generator torque in Nm\n'.format(int(rosco_vt['Ind_GenTq'])))
    file.write('{0:<12d}        ! Ind_YawRate       - The column in OL_Filename that contains the generator torque in Nm\n'.format(int(rosco_vt['Ind_YawRate'])))
    file.close()
    
    # Write Open loop input
    if rosco_vt['OL_Mode'] and hasattr(controller, 'OpenLoop'):
        write_ol_control(controller)

def read_DISCON(DISCON_filename):
    '''
    Read the DISCON input file.

    Parameters:
    ----------
    DISCON_filename: string
        Name of DISCON input file to read
    
    Returns:
    --------
    DISCON_in: Dict
        Dictionary containing input parameters from DISCON_in, organized by parameter name
    '''
    
    DISCON_in = {}
    with open(DISCON_filename) as discon:
        for line in discon:

            # Skip whitespace and comment lines
            if (line[0] != '!') == (len(line.strip()) != 0):
                
                if (line.split()[1] != '!'):    # Array valued entries
                    array_length = line.split().index('!')
                    param = line.split()[array_length+1]
                    values = np.array( [float(x) for x in line.split()[:array_length]] )
                    DISCON_in[param] = values
                else:                           # All other entries
                    param = line.split()[2]
                    value = line.split()[0]
                    # Remove printed quotations if string is in quotes
                    if (value[0] == '"') or (value[0] == "'"):
                        value = value[1:-1]
                    else:
                        value = float(value)
                    DISCON_in[param] = value

    return DISCON_in
    

def write_rotor_performance(turbine,txt_filename='Cp_Ct_Cq.txt'):
    '''
    Write text file containing rotor performance data

    Parameters:
    ------------
        txt_filename: str, optional
                        Desired output filename to print rotor performance data. Default is Cp_Ct_Cq.txt
    '''
    print('Writing rotor performance text file: {}'.format(txt_filename))
    file = open(txt_filename,'w')
    # Headerlines
    file.write('# ----- Rotor performance tables for the {} wind turbine ----- \n'.format(turbine.TurbineName))
    file.write('# ------------ Written on {} using the ROSCO toolbox ------------ \n\n'.format(now.strftime('%b-%d-%y')))

    # Pitch angles, TSR, and wind speed
    file.write('# Pitch angle vector, {} entries - x axis (matrix columns) (deg)\n'.format(len(turbine.Cp.pitch_initial_rad)))
    for i in range(len(turbine.Cp.pitch_initial_rad)):
        file.write('{:0.4}   '.format(turbine.Cp.pitch_initial_rad[i] * rad2deg))
    file.write('\n# TSR vector, {} entries - y axis (matrix rows) (-)\n'.format(len(turbine.TSR_initial)))
    for i in range(len(turbine.TSR_initial)):
        file.write('{:0.4}    '.format(turbine.Cp.TSR_initial[i]))
    file.write('\n# Wind speed vector - z axis (m/s)\n')
    file.write('{:0.4}    '.format(turbine.v_rated))
    file.write('\n')
    
    # Cp
    file.write('\n# Power coefficient\n\n')
    for i in range(len(turbine.Cp.TSR_initial)):
        for j in range(len(turbine.Cp.pitch_initial_rad)):
            file.write('{0:.6f}   '.format(turbine.Cp_table[i,j]))
        file.write('\n')
    file.write('\n')
    
    # Ct
    file.write('\n#  Thrust coefficient\n\n')
    for i in range(len(turbine.Ct.TSR_initial)):
        for j in range(len(turbine.Ct.pitch_initial_rad)):
            file.write('{0:.6f}   '.format(turbine.Ct_table[i,j]))
        file.write('\n')
    file.write('\n')
    
    # Cq
    file.write('\n# Torque coefficient\n\n')
    for i in range(len(turbine.Cq.TSR_initial)):
        for j in range(len(turbine.Cq.pitch_initial_rad)):
            file.write('{0:.6f}   '.format(turbine.Cq_table[i,j]))
        file.write('\n')
    file.write('\n')
    file.close()


def load_from_txt(txt_filename):
    '''
    Load rotor performance data from a *.txt file. 

    Parameters:
    -----------
        txt_filename: str
                        Filename of the text containing the Cp, Ct, and Cq data. This should be in the format printed by the write_rotorperformance function
    '''
    print('Loading rotor performace data from text file:', txt_filename)

    with open(txt_filename) as pfile:
        for line in pfile:
            # Read Blade Pitch Angles (degrees)
            if 'Pitch angle' in line:
                pitch_initial = np.array([float(x) for x in pfile.readline().strip().split()])
                pitch_initial_rad = pitch_initial * deg2rad             # degrees to rad            -- should this be conditional?

            # Read Tip Speed Ratios (rad)
            if 'TSR' in line:
                TSR_initial = np.array([float(x) for x in pfile.readline().strip().split()])
            
            # Read Power Coefficients
            if 'Power' in line:
                pfile.readline()
                Cp = np.empty((len(TSR_initial),len(pitch_initial)))
                for tsr_i in range(len(TSR_initial)):
                    Cp[tsr_i] = np.array([float(x) for x in pfile.readline().strip().split()])
            
            # Read Thrust Coefficients
            if 'Thrust' in line:
                pfile.readline()
                Ct = np.empty((len(TSR_initial),len(pitch_initial)))
                for tsr_i in range(len(TSR_initial)):
                    Ct[tsr_i] = np.array([float(x) for x in pfile.readline().strip().split()])

            # Read Torque Coefficients
            if 'Torque' in line:
                pfile.readline()
                Cq = np.empty((len(TSR_initial),len(pitch_initial)))
                for tsr_i in range(len(TSR_initial)):
                    Cq[tsr_i] = np.array([float(x) for x in pfile.readline().strip().split()])

        # return pitch_initial_rad TSR_initial Cp Ct Cq
        # Store necessary metrics for analysis and tuning
        # self.pitch_initial_rad = pitch_initial_rad
        # self.TSR_initial = TSR_initial
        # self.Cp_table = Cp
        # self.Ct_table = Ct 
        # self.Cq_table = Cq
        return pitch_initial_rad, TSR_initial, Cp, Ct, Cq


def DISCON_dict(turbine, controller, txt_filename=None):
    '''
    Convert the turbine and controller objects to a dictionary organized by the parameter names 
    that are defined in the DISCON.IN file.

    Parameters
    ----------
    turbine: obj
        Turbine object output from the turbine class
    controller: obj
        Controller object output from the controller class
    txt_filename: string, optional
        Name of rotor performance filename
    '''
    DISCON_dict = {}
    # ------- DEBUG -------
    DISCON_dict['LoggingLevel']	    = int(controller.LoggingLevel)
    # ------- CONTROLLER FLAGS -------
    DISCON_dict['F_LPFType']	    = int(controller.F_LPFType)
    DISCON_dict['F_NotchType']		= int(controller.F_NotchType)
    DISCON_dict['IPC_ControlMode']	= int(controller.IPC_ControlMode)
    DISCON_dict['VS_ControlMode']	= int(controller.VS_ControlMode)
    DISCON_dict['PC_ControlMode']   = int(controller.PC_ControlMode)
    DISCON_dict['Y_ControlMode']	= int(controller.Y_ControlMode)
    DISCON_dict['SS_Mode']          = int(controller.SS_Mode)
    DISCON_dict['WE_Mode']          = int(controller.WE_Mode)
    DISCON_dict['PS_Mode']          = int(controller.PS_Mode > 0)
    DISCON_dict['SD_Mode']          = int(controller.SD_Mode)
    DISCON_dict['Fl_Mode']          = int(controller.Fl_Mode)
    DISCON_dict['Flp_Mode']         = int(controller.Flp_Mode)
    DISCON_dict['OL_Mode']          = int(controller.OL_Mode)
    # ------- FILTERS -------
    DISCON_dict['F_LPFCornerFreq']	    = turbine.bld_edgewise_freq * 1/4
    DISCON_dict['F_LPFDamping']		    = controller.F_LPFDamping
    if controller.Flp_Mode > 0:
        DISCON_dict['F_NotchCornerFreq'] = turbine.bld_flapwise_freq
        DISCON_dict['F_NotchBetaNumDen'] = [0.0, 0.5]
    else:
        DISCON_dict['F_NotchCornerFreq'] = controller.twr_freq
        DISCON_dict['F_NotchBetaNumDen'] = [0.0, 0.25]
    DISCON_dict['F_WECornerFreq'] = controller.f_we_cornerfreq
    DISCON_dict['F_SSCornerFreq'] = controller.f_ss_cornerfreq
    DISCON_dict['F_FlHighPassFreq'] = controller.f_fl_highpassfreq
    DISCON_dict['F_FlCornerFreq'] = [controller.ptfm_freq, 1.0]
    DISCON_dict['F_FlpCornerFreq'] = [turbine.bld_flapwise_freq*3, 1.0]
    DISCON_dict['F_WECornerFreq']       = controller.f_we_cornerfreq
    DISCON_dict['F_SSCornerFreq']       = controller.f_ss_cornerfreq
    DISCON_dict['F_FlHighPassFreq']     = controller.f_fl_highpassfreq
    DISCON_dict['F_FlCornerFreq']       = [controller.ptfm_freq, 1.0]
    # ------- BLADE PITCH CONTROL -------
    DISCON_dict['PC_GS_n']			= len(controller.pitch_op_pc)
    DISCON_dict['PC_GS_angles']	    = controller.pitch_op_pc
    DISCON_dict['PC_GS_KP']		    = controller.pc_gain_schedule.Kp
    DISCON_dict['PC_GS_KI']		    = controller.pc_gain_schedule.Ki
    DISCON_dict['PC_GS_KD']			= [0.0 for i in range(len(controller.pc_gain_schedule.Ki))]
    DISCON_dict['PC_GS_TF']			= [0.0 for i in range(len(controller.pc_gain_schedule.Ki))]
    DISCON_dict['PC_MaxPit']		= controller.max_pitch
    DISCON_dict['PC_MinPit']		= controller.min_pitch
    DISCON_dict['PC_MaxRat']		= turbine.max_pitch_rate
    DISCON_dict['PC_MinRat']		= turbine.min_pitch_rate
    DISCON_dict['PC_RefSpd']		= turbine.rated_rotor_speed*turbine.Ng
    DISCON_dict['PC_FinePit']		= controller.min_pitch
    DISCON_dict['PC_Switch']		= 1 * deg2rad
    # ------- INDIVIDUAL PITCH CONTROL -------
    DISCON_dict['IPC_IntSat']		= 0.087266
    DISCON_dict['IPC_KP']           = [controller.Kp_ipc1p, controller.Kp_ipc2p]
    DISCON_dict['IPC_KI']           = [controller.Ki_ipc1p, controller.Ki_ipc2p]
    DISCON_dict['IPC_aziOffset']	= [0.0, 0.0]
    DISCON_dict['IPC_CornerFreqAct'] = 0.0
    # ------- VS TORQUE CONTROL -------
    DISCON_dict['VS_GenEff']		= turbine.GenEff
    DISCON_dict['VS_ArSatTq']		= turbine.rated_torque
    DISCON_dict['VS_MaxRat']		= turbine.max_torque_rate
    DISCON_dict['VS_MaxTq']			= turbine.max_torque
    DISCON_dict['VS_MinTq']			= 0.0
    DISCON_dict['VS_MinOMSpd']		= controller.vs_minspd  * turbine.Ng
    DISCON_dict['VS_Rgn2K']			= controller.vs_rgn2K
    DISCON_dict['VS_RtPwr']			= turbine.rated_power
    DISCON_dict['VS_RtTq']			= turbine.rated_torque
    DISCON_dict['VS_RefSpd']		= controller.vs_refspd
    DISCON_dict['VS_n']				= 1
    DISCON_dict['VS_KP']			= controller.vs_gain_schedule.Kp[-1]
    DISCON_dict['VS_KI']			= controller.vs_gain_schedule.Ki[-1]
    DISCON_dict['VS_TSRopt']		= turbine.TSR_operational
    # ------- SETPOINT SMOOTHER -------
    DISCON_dict['SS_VSGain']         = controller.ss_vsgain
    DISCON_dict['SS_PCGain']         = controller.ss_pcgain
    # ------- WIND SPEED ESTIMATOR -------
    DISCON_dict['WE_BladeRadius']	= turbine.rotor_radius
    DISCON_dict['WE_CP_n']			= 1
    DISCON_dict['WE_CP']            = 0
    DISCON_dict['WE_Gamma']			= 0.0
    DISCON_dict['WE_GearboxRatio']	= turbine.Ng
    DISCON_dict['WE_Jtot']			= turbine.J
    DISCON_dict['WE_RhoAir']		= turbine.rho
    DISCON_dict['PerfFileName']     = txt_filename
    DISCON_dict['PerfTableSize']    = [len(turbine.Cp.pitch_initial_rad),len(turbine.Cp.TSR_initial)]
    DISCON_dict['WE_FOPoles_N']     = len(controller.A)
    DISCON_dict['WE_FOPoles_v']     = controller.v
    DISCON_dict['WE_FOPoles']       = controller.A
    # ------- YAW CONTROL -------
    DISCON_dict['Y_ErrThresh']		= 0.1396
    DISCON_dict['Y_IPC_IntSat']		= 0.0
    DISCON_dict['Y_IPC_n']			= 1
    DISCON_dict['Y_IPC_KP']			= 0.0
    DISCON_dict['Y_IPC_KI']			= 0.0
    DISCON_dict['Y_IPC_omegaLP']    = 0.2094
    DISCON_dict['Y_IPC_zetaLP']		= 1
    DISCON_dict['Y_MErrSet']		= 0.0
    DISCON_dict['Y_omegaLPFast']	= 0.2094
    DISCON_dict['Y_omegaLPSlow']	= 0.1047
    DISCON_dict['Y_Rate']			= 0.0052
    # ------- TOWER FORE-AFT DAMPING -------
    DISCON_dict['FA_KI']                = -1
    DISCON_dict['FA_HPFCornerFreq'] = 0.0
    DISCON_dict['FA_IntSat']		 = 0.0
    # ------- MINIMUM PITCH SATURATION -------
    DISCON_dict['PS_BldPitchMin_N'] = len(controller.ps_min_bld_pitch)
    DISCON_dict['PS_WindSpeeds']    = controller.v
    DISCON_dict['PS_BldPitchMin']   = controller.ps_min_bld_pitch
    # ------- SHUTDOWN -------
    DISCON_dict['SD_MaxPit']        = controller.sd_maxpit
    DISCON_dict['SD_CornerFreq']    = controller.f_sd_cornerfreq
    # ------- Floating -------
    DISCON_dict['Fl_Kp']            = controller.Kp_float
    # ------- FLAP ACTUATION -------
    DISCON_dict['Flp_Angle']        = controller.flp_angle
    DISCON_dict['Flp_Kp']           = controller.Kp_flap[-1]
    DISCON_dict['Flp_Ki']           = controller.Ki_flap[-1]
    DISCON_dict['Flp_MaxPit']       = controller.flp_maxpit
    # ------- Open Loop Control -------
    DISCON_dict['OL_Filename']     = controller.OL_Filename
    DISCON_dict['Ind_Breakpoint']  = controller.OL_Ind_Breakpoint
    DISCON_dict['Ind_BldPitch']    = controller.OL_Ind_BldPitch
    DISCON_dict['Ind_GenTq']       = controller.OL_Ind_GenTq
    DISCON_dict['Ind_YawRate']     = controller.OL_Ind_YawRate

    return DISCON_dict


def run_openfast(fast_dir, fastcall='openfast', fastfile=None, chdir=True, restart=False):
    '''
    Runs a openfast openfast simulation.
    
    NOTE: Enabling chdir can help reduce file path errors in the DISCON.IN file.

    Parameters:
    ------------
        fast_dir: string
                Name of OpenFAST directory containing input files.
        fast_file: string
                Name of OpenFAST directory containing input files.
        fastcall: string, optional
                Line used to call openfast when executing from the terminal.
        fastfile: string, optional
                Filename for *.fst input file. Function will find *.fst if not provided.
        chdir: bool, optional
                Change directory to openfast model directory before running.
    '''

    # Define OpenFAST input filename
    if not fastfile:
        for file in os.listdir(fast_dir):
            if file.endswith('.fst'):
                fastfile = file
    print('Using {} to run OpenFAST simulation'.format(fastfile))

    if chdir: # Change cwd before calling OpenFAST -- note: This is an artifact of needing to call OpenFAST from the same directory as DISCON.IN
        cwd = fast_dir
    else:
        cwd = None

    print('Running OpenFAST simulation for {} through the ROSCO toolbox...'.format(fastfile))
        # os.system('{} {}'.format(fastcall, os.path.join(fastfile)))
    if restart:
        subprocess.run([fastcall,'-restart', os.path.join(fastfile)], check=True, cwd=cwd)
    else:
        subprocess.run([fastcall, os.path.join(fastfile)], check=True, cwd=cwd)
    print('OpenFAST simulation complete.')


def list_check(x, return_bool=True):
    '''
    Check if the input is list-like or not

    Parameters:
    -----------
        x: int, float, list, or np.ndarray
            input to check
        return_bool: bool
            if true, returns True or False 

    '''
    if isinstance(x, (int, float)):
        y = x
        is_list = False
    elif isinstance(x, list):
        if len(x) == 1:
            y = x[0]
            is_list = False
        else:
            y = x
            is_list = True
    elif isinstance(x, np.ndarray):
        if x.size == 1:
            is_list = False
            y = float(x)
        else:
            is_list = True
            y = x
    else:
        raise AttributeError('Cannot run list_check for variable of type: {}'.format(type(x)))

    if return_bool:
        return is_list
    else:
        return y



'''
Utilities to aid linear model controller development and analysis
'''
import numpy as np
import scipy as sp
import control as co
import warnings
import copy

def pc_openloop(linturb, controller, u):
    '''
    This function creates a set of openloop transfer function for the wind turbine with a pitch controller
    
    Inputs
    ------
    linturb: object
        LinearTurbineModel object
    controller: object
        ROSCO toolbox controller object
    u: float
        wind speed to evaluate system at
    '''
    if linturb.B_ops.shape[1] > 1:
        raise TypeError(
            'OpenLoop system calculations are only supported for single input systems, please run LinearTurbine.trim_system with len(desInputs) = 1')
    if linturb.C_ops.shape[0] > 1:
        raise TypeError(
            'OpenLoop system calculations are only supported for single output systems, please run LinearTurbine.trim_system with len(desOutputs) = 1')

    # Interpolate plant and controller
    P = interp_plant(linturb, u)
    Cs = interp_pitch_controller(controller, u)

    # Combine controller and plant
    sys_ol = Cs*P

    return sys_ol


def pc_closedloop(linturb, controller, u):
    '''
    This function creates a set of closed loop transfer function for the wind turbine with a pitch controller
    
    Inputs
    ------
    linturb: object
        LinearTurbineModel object
    controller: object
        ROSCO toolbox controller object
    u: float
        wind speed to evaluate closed loop system at
    Returns
    -------
    sys_cl: scipy StateSpaceContinuous 
        closed loop system
    '''
    if linturb.B_ops.shape[1] > 1:
        raise TypeError(
            'OpenLoop system calculations are only supported for single input systems, please run LinearTurbine.trim_system with len(desInputs) = 1')
    if linturb.C_ops.shape[0] > 1:
        raise TypeError(
            'OpenLoop system calculations are only supported for single output systems, please run LinearTurbine.trim_system with len(desOutputs) = 1')

    # Interpolate plant and controller
    P = interp_plant(linturb, u)
    Cs = interp_pitch_controller(controller, u)

    # Combine controller and plant
    sys_cl = feedback(1, Cs*P)

    return sys_cl


def pc_sensitivity(linturb, controller, u):
    '''
    This function finds the sensitivity function for the wind turbine with a pitch controller
    
    Inputs
    ------
    linturb: object
        LinearTurbineModel object
    controller: object
        ROSCO toolbox controller object
    u: float
        wind speed to evaluate system at
    '''

    if linturb.B_ops.shape[1] > 1:
        raise TypeError(
            'OpenLoop system calculations are only supported for single input systems, please run LinearTurbine.trim_system with len(desInputs) = 1')
    if linturb.C_ops.shape[0] > 1:
        raise TypeError(
            'OpenLoop system calculations are only supported for single output systems, please run LinearTurbine.trim_system with len(desOutputs) = 1')

    # Interpolate plant and controller
    P = interp_plant(linturb, u)
    Cs = interp_pitch_controller(controller, u)

    # Calculate sensitivity function
    sens = feedback(1, Cs*P)

    return sens

def smargin(linturb, controller, u_eval):
    # try:
    #     k_float = inputs[1]
    # except:
    sens_sys = pc_sensitivity(linturb, controller, u_eval)
    ol_sys = pc_openloop(linturb, controller, u_eval)

    sp_plant = sp.signal.StateSpace(ol_sys.A, ol_sys.B, ol_sys.C, ol_sys.D)
    sp_sens = sp.signal.StateSpace(sens_sys.A, sens_sys.B, sens_sys.C, sens_sys.D)

    def nyquist_min(om): return np.abs(sp.signal.freqresp(sp_plant, w=om)[1] + 1.)
    def sens_min(om): return -sp.signal.bode(sp_sens, w=om)[1]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # Find first local maxima in sensitivity function
        ws, m, _ = sp.signal.bode(sp_sens, n=100000)
        m0 = m[0]
        m1 = m[1]
        i = 1
        w0 = ws[0]
        while m1 >= m0 and i < len(ws)-1:
            m0 = m1
            m1 = m[i+1]
            i += 1
            w0 = ws[i]

        sm_mag = sp.signal.freqresp(sp_plant, w=w0)[1]
        sm = np.sqrt((1 - np.abs(sm_mag.real))**2 + sm_mag.imag**2)
        nearest_nyquist = nyquist_min(ws).min()
        nearest_nyquist_freq = ws[nyquist_min(ws).argmin()]
        mag_at_min = sp.signal.freqresp(sp_plant, w=nearest_nyquist_freq)[1]
        if any(sp_sens.poles > 0):
            if nearest_nyquist < sm:
                res = sp.optimize.minimize(nyquist_min, nearest_nyquist_freq, method='SLSQP', options={
                    'finite_diff_rel_step': 1e-8})
                sm2 = min(abs(res.fun), abs(nearest_nyquist))

                sm_list = [sm, sm2]
                mag_list = [np.abs(sm_mag), np.abs(mag_at_min)]
                sm = sm_list[np.argmax(mag_list)]
            sm *= -1  # Flip sign because it's unstable
        else:
            res = sp.optimize.minimize(nyquist_min, nearest_nyquist_freq, method='SLSQP',
                                       options={'finite_diff_rel_step': 1e-6})
            sm = min(res.fun, nearest_nyquist)



    return sm

def interp_plant(linturb, v, return_scipy=True):
    '''
    Interpolate linear turbine plant 

    Inputs
    ------
    linturb: object
        LinearTurbineModel object
    v: float
        wind speed to interpolate linturb at
    return_scipy: bool, optional
        True:  return type is scipy's StateSpaceContinuous
        False: return type is python control toolbox's StateSpace

    Returns
    -------
    P: StateSpaceContinous or StateSpace
    '''

    # Find interpolated plant on v
    Ap = interp_matrix(linturb.u_h, linturb.A_ops, v)
    Bp = interp_matrix(linturb.u_h, linturb.B_ops, v)
    Cp = interp_matrix(linturb.u_h, linturb.C_ops, v)
    Dp = interp_matrix(linturb.u_h, linturb.D_ops, v)

    if return_scipy:
        P = sp.signal.StateSpace(Ap, Bp, Cp, Dp)
    else:
        P = co.StateSpace(Ap, Bp, Cp, Dp)

    return P


def interp_pitch_controller(controller, v, return_scipy=True):
    '''
    Interpolate ROSCO toolbox pitch controller and return linear model

    Inputs
    ------
    controller: object
        ROSCO toolbox controller object
    v: float
        wind speed to interpolate linturb at
    return_scipy: bool, optional
        True:  return type is scipy's StateSpaceContinuous
        False: return type is python control toolbox's StateSpace

    Returns
    -------
    P: StateSpaceContinous or StateSpace
    '''

    # interpolate controller on v
    pitch = interp_matrix(controller.v, controller.pitch_op, v)
    kp = float(interp_matrix(controller.pitch_op_pc, controller.pc_gain_schedule.Kp, pitch))
    ki = float(interp_matrix(controller.pitch_op_pc, controller.pc_gain_schedule.Ki, pitch))
    if return_scipy:
        A, B, C, D = sp.signal.tf2ss([kp, ki], [1, 0])
        Cs = sp.signal.StateSpace(A, B, C, D)
    else:
        Cs = co.tf2ss(co.TransferFunction([kp, ki], [1, 0]))

    return Cs


def add_pcomp(linturb, k_float):
    '''
    Incorporate the effect of a parallel compensation term to a linear turbine model.

    Inputs
    ------
    linturb: object
        LinearTurbineModel object
    k_float: float
        parallel compensation feedback gain

    Returns
    -------
    linturb2: object
        Modified linturb with parallel compensation term included
    '''

    state_str = 'derivative of 1st tower fore-aft'
    state_idx = np.flatnonzero(np.core.defchararray.find(
        linturb.DescStates, state_str) > -1).tolist()
    input_str = 'collective blade-pitch'
    input_idx = np.flatnonzero(np.core.defchararray.find(
        linturb.DescCntrlInpt, input_str) > -1).tolist()

    K = np.zeros((linturb.B_ops.shape[1], linturb.A_ops.shape[1], linturb.A_ops.shape[2]))
    K[input_idx, state_idx, :] = -k_float # NJA: negative to account of OF linearization sign conventions

    linturb2 = copy.copy(linturb)
    linturb2.A_ops = linturb.A_ops + linturb.B_ops * K

    return linturb2


def interp_matrix(x, matrix_3D, q):
    q = np.clip(q, x.min(), x.max())
    f_m = sp.interpolate.interp1d(x, matrix_3D)

    return f_m(np.squeeze(q))


def _convert_to_ss(sys):
    if isinstance(sys, (int, float, complex, np.float)):
        ss = sp.signal.StateSpace([0], [0], [0], sys)
    else:
        ss = sys

    return ss


def feedback(sys1, sys2, sign=-1):
    """
    Feedback interconnection between two LTI systems.

    This underlying methods here are pulled from the python 
    control toolbox, but made to work with scipy state space 
    objects.
    """
    sys1 = _convert_to_ss(sys1)
    sys2 = _convert_to_ss(sys2)

    # Check to make sure the dimensions are OK
    if (sys1.inputs != sys2.outputs) or (sys1.outputs != sys2.inputs):
        raise ValueError("State space systems don't have compatible "
                         "inputs/outputs for feedback.")
    if sys1.dt != sys2.dt:
        raise ValueError(
            "Both state space systems must have the same timestep (dt), or both be continuous time.")
    else:
        dt = sys1.dt

    A1 = sys1.A
    B1 = sys1.B
    C1 = sys1.C
    D1 = sys1.D
    A2 = sys2.A
    B2 = sys2.B
    C2 = sys2.C
    D2 = sys2.D

    F = np.eye(sys1.inputs) - sign * np.dot(D2, D1)
    if np.linalg.matrix_rank(F) != sys1.inputs:
        raise ValueError("I - sign * D2 * D1 is singular to working precision.")

    # Precompute F\D2 and F\C2 (E = inv(F))
    # We can solve two linear systems in one pass, since the
    # coefficients matrix F is the same. Thus, we perform the LU
    # decomposition (cubic runtime complexity) of F only once!
    # The remaining back substitutions are only quadratic in runtime.
    E_D2_C2 = np.linalg.solve(F, np.concatenate((D2, C2), axis=1))
    E_D2 = E_D2_C2[:, :sys2.inputs]
    E_C2 = E_D2_C2[:, sys2.inputs:]

    T1 = np.eye(sys1.outputs) + sign * np.dot(D1, E_D2)
    T2 = np.eye(sys1.inputs) + sign * np.dot(E_D2, D1)

    A = np.concatenate(
        (np.concatenate(
            (A1 + sign * np.dot(np.dot(B1, E_D2), C1),
             sign * np.dot(B1, E_C2)), axis=1),
         np.concatenate(
             (np.dot(B2, np.dot(T1, C1)),
              A2 + sign * np.dot(np.dot(B2, D1), E_C2)), axis=1)),
        axis=0)
    B = np.concatenate((np.dot(B1, T2), np.dot(np.dot(B2, D1), T2)), axis=0)
    C = np.concatenate((np.dot(T1, C1), sign * np.dot(D1, E_C2)), axis=1)
    D = np.dot(D1, T2)

    if dt:
        return sp.signal.StateSpace(A, B, C, D, dt=dt)
    else:
        return sp.signal.StateSpace(A, B, C, D)



import numpy as np
import os, sys
# import matplotlib.pyplot as plt 

from ROSCO_toolbox.ofTools.fast_io.turbsim_io.turbsim_writer import TurbsimBuilder
from ROSCO_toolbox.ofTools.fast_io.turbsim_io.turbsim_wrapper import Turbsim_wrapper
from ROSCO_toolbox.ofTools.fast_io.turbsim_io.turbsim_vartrees import turbsiminputs

# from AeroelasticSE.Turbsim_mdao.pyturbsim_wrapper import pyTurbsim_wrapper

class pyIECWind_extreme():

    def __init__(self):

        self.Turbine_Class    = 'I'    # IEC Wind Turbine Class
        self.Turbulence_Class = 'B'    # IEC Turbulance Class
        self.Vert_Slope       = 0      # Vertical slope of the wind inflow (deg)
        self.TStart           = 30     # Time to start transient conditions (s)
        self.dt               = 0.05   # Transient wind time step (s)
        self.dir_change       = 'both' # '+','-','both': sign for transient events in EDC, EWS
        self.shear_orient     = 'both' # 'v','h','both': vertical or horizontal shear for EWS
        self.z_hub            = 90.    # wind turbine hub height (m)
        self.D                = 126.   # rotor diameter (m)
        
        self.T0               = 0.
        self.TF               = 630.

    def setup(self):
        # General turbulence parameters: 6.3
        # Sigma_1: logitudinal turbulence scale parameter

        # Setup
        if self.Turbine_Class == 'I':
            self.V_ref = 50.
        elif self.Turbine_Class == 'II':
            self.V_ref = 42.5
        elif self.Turbine_Class == 'III':
            self.V_ref = 37.5
        elif self.Turbine_Class == 'IV':
            self.V_ref = 30.
        self.V_ave = self.V_ref*0.2

        if self.Turbulence_Class == 'A+':
            self.I_ref = 0.18
        elif self.Turbulence_Class == 'A':
            self.I_ref = 0.16
        elif self.Turbulence_Class == 'B':
            self.I_ref = 0.14
        elif self.Turbulence_Class == 'C':
            self.I_ref = 0.12

        if self.z_hub > 60:
            self.Sigma_1 = 42
        else:
            self.Sigma_1 = 0.7*self.z_hub
            

    def NTM(self, V_hub):
        # Normal turbulence model: 6.3.1.3
        b = 5.6
        sigma_1 = self.I_ref*(0.75*V_hub + b)
        return sigma_1

    def ETM(self, V_hub):
        # Extreme turbulence model: 6.3.2.3
        c = 2
        sigma_1 = c*self.I_ref*(0.072*(self.V_ave/c + 3)*(V_hub/c - 4) + 10)
        return sigma_1

    def EWM(self, V_hub):
        # Extreme wind speed model: 6.3.2.1
                
        # Steady
        V_e50 = 1.4*self.V_ref
        V_e1 = 0.8*V_e50
        # Turb
        V_50 = self.V_ref
        V_1 = 0.8*V_50
        sigma_1 = 0.11*V_hub

        return sigma_1, V_e50, V_e1, V_50, V_1

    def EOG(self, V_hub_in):
        # Extreme operating guest: 6.3.2.2

        self.setup()

        T = 10.5
        t = np.linspace(0., T, num=(T/self.dt+1))

        # constants from standard
        alpha = 0.2

        # Flow angle adjustments
        V_hub = V_hub_in*np.cos(self.Vert_Slope*np.pi/180)
        V_vert_mag = V_hub_in*np.sin(self.Vert_Slope*np.pi/180)

        sigma_1 = self.NTM(V_hub)
        __, __, V_e1, __, __ = self.EWM(V_hub)

        # Contant variables
        V = np.zeros_like(t)+V_hub
        V_dir = np.zeros_like(t)
        V_vert = np.zeros_like(t)+V_vert_mag
        shear_horz = np.zeros_like(t)
        shear_vert = np.zeros_like(t)+alpha
        shear_vert_lin = np.zeros_like(t)
        V_gust = np.zeros_like(t)
        upflow = np.zeros_like(t)

        V_gust = min([ 1.35*(V_e1 - V_hub), 3.3*(sigma_1/(1+0.1*(self.D/self.Sigma_1))) ])

        V_gust_t = np.zeros_like(t)
        for i, ti in enumerate(t):
            if ti<T:
                V_gust_t[i] = 0. - 0.37*V_gust*np.sin(3*np.pi*ti/T)*(1-np.cos(2*np.pi*ti/T))
            else:
                V_gust_t[i] = 0.

        # Write Files
        self.fname_out = []
        self.fname_type = []
        fname = self.case_name + '_EOG_U%2.1f.wnd'%V_hub_in
        data = np.column_stack((t, V, V_dir, V_vert, shear_horz, shear_vert, shear_vert_lin, V_gust_t, upflow))
        hd = []
        hd.append('! Extreme operating guest\n')
        hd = self.heading_common(hd)
        hd = self.heading_variable(hd, V_hub_in=V_hub_in, sigma_1=sigma_1)
        self.write_wnd(fname, data, hd)
        self.fname_out.append(os.path.realpath(os.path.normpath(os.path.join(self.outdir, fname))))
        self.fname_type.append(2)



    def EDC(self, V_hub_in):
        # Extreme direction change: 6.3.2.4

        self.setup()
        
        T = 6.
        t = np.linspace(0., T, num=(T/self.dt+1))

        # constants from standard
        alpha = 0.2

        # Flow angle adjustments
        V_hub = V_hub_in*np.cos(self.Vert_Slope*np.pi/180)
        V_vert_mag = V_hub_in*np.sin(self.Vert_Slope*np.pi/180)

        sigma_1 = self.NTM(V_hub)

        # Contant variables
        V = np.zeros_like(t)+V_hub
        V_vert = np.zeros_like(t)+V_vert_mag
        shear_horz = np.zeros_like(t)
        shear_vert = np.zeros_like(t)+alpha
        shear_vert_lin = np.zeros_like(t)
        V_gust = np.zeros_like(t)
        upflow = np.zeros_like(t)

        # Transient
        Theta_e = 4.*np.arctan(sigma_1/(V_hub*(1.+0.01*(self.D/self.Sigma_1))))*180./np.pi
        if Theta_e > 180.:
            Theta_e = 180.

        Theta_p = np.zeros_like(t)
        Theta_n = np.zeros_like(t)
        for i, ti in enumerate(t):
            if ti<T:
                Theta_p[i] = 0.5*Theta_e*(1-np.cos(np.pi*ti/T))
                Theta_n[i] = -1*0.5*Theta_e*(1-np.cos(np.pi*ti/T))
            else:
                Theta_p[i] = Theta_e
                Theta_n[i] = -1*Theta_e

        # Write Files
        self.fname_out = []
        self.fname_type = []
        if self.dir_change.lower() == 'both' or self.dir_change.lower() == '+':
            ## Vert
            fname = self.case_name + '_EDC_P_U%2.1f.wnd'%V_hub_in
            data = np.column_stack((t, V, Theta_p, V_vert, shear_horz, shear_vert, shear_vert_lin, V_gust, upflow))
            hd = []
            hd.append('! Exteme Vertical Wind Shear, positive\n')
            hd = self.heading_common(hd)
            hd = self.heading_variable(hd, V_hub_in=V_hub_in, sigma_1=sigma_1)
            self.write_wnd(fname, data, hd)
            self.fname_out.append(os.path.realpath(os.path.normpath(os.path.join(self.outdir, fname))))
            self.fname_type.append(2)


        if self.dir_change.lower() == 'both' or self.dir_change.lower() == '-':
            ## Vert
            fname = self.case_name + '_EDC_N_U%2.1f.wnd'%V_hub_in
            data = np.column_stack((t, V, Theta_n, V_vert, shear_horz, shear_vert, shear_vert_lin, V_gust, upflow))
            hd = []
            hd.append('! Exteme Vertical Wind Shear, negative\n')
            hd = self.heading_common(hd)
            hd = self.heading_variable(hd, V_hub_in=V_hub_in, sigma_1=sigma_1)
            self.write_wnd(fname, data, hd)
            self.fname_out.append(os.path.realpath(os.path.normpath(os.path.join(self.outdir, fname))))
            self.fname_type.append(2)

         
    def ECD(self, V_hub_in):
        # Extreme coherent gust with direction change: 6.3.2.5

        self.setup()
        
        T = 10.
        t = np.linspace(0., T, num=int((T/self.dt+1)))

        # constants from standard
        alpha = 0.2
        V_cg = 15 #m/s

        # Flow angle adjustments
        V_hub = V_hub_in*np.cos(self.Vert_Slope*np.pi/180)
        V_vert_mag = V_hub_in*np.sin(self.Vert_Slope*np.pi/180)

        sigma_1 = self.NTM(V_hub)

        # Contant variables
        V_vert = np.zeros_like(t)+V_vert_mag
        shear_horz = np.zeros_like(t)
        shear_vert = np.zeros_like(t)+alpha
        shear_vert_lin = np.zeros_like(t)
        V_gust = np.zeros_like(t)
        upflow = np.zeros_like(t)

        # Transient
        if V_hub < 4:
            Theta_cg = 180
        else:
            Theta_cg = 720/V_hub

        Theta_p = np.zeros_like(t)
        Theta_n = np.zeros_like(t)
        V = np.zeros_like(t)
        for i, ti in enumerate(t):
            if ti<T:
                V[i] = V_hub + 0.5*V_cg*(1-np.cos(np.pi*ti/T))
                Theta_p[i] = 0.5*Theta_cg*(1-np.cos(np.pi*ti/T))
                Theta_n[i] = -1*0.5*Theta_cg*(1-np.cos(np.pi*ti/T))
            else:
                V[i] = V_hub+V_cg
                Theta_p[i] = Theta_cg
                Theta_n[i] = -1*Theta_cg

        # Write Files
        self.fname_out = []
        self.fname_type = []
        if self.dir_change.lower() == 'both' or self.dir_change.lower() == '+':
            ## Vert
            fname = self.case_name + '_ECD_P_U%2.1f.wnd'%V_hub_in
            data = np.column_stack((t, V, Theta_p, V_vert, shear_horz, shear_vert, shear_vert_lin, V_gust, upflow))
            hd = []
            hd.append('! Exteme coherent gust with direction change, positive\n')
            hd = self.heading_common(hd)
            hd = self.heading_variable(hd, V_hub_in=V_hub_in)
            self.write_wnd(fname, data, hd)
            self.fname_out.append(os.path.realpath(os.path.normpath(os.path.join(self.outdir, fname))))
            self.fname_type.append(2)


        if self.dir_change.lower() == 'both' or self.dir_change.lower() == '-':
            ## Vert
            fname = self.case_name + '_ECD_N_U%2.1f.wnd'%V_hub_in
            data = np.column_stack((t, V, Theta_n, V_vert, shear_horz, shear_vert, shear_vert_lin, V_gust, upflow))
            hd = []
            hd.append('! Exteme coherent gust with direction change, negative\n')
            hd = self.heading_common(hd)
            hd = self.heading_variable(hd, V_hub_in=V_hub_in)
            self.write_wnd(fname, data, hd)
            self.fname_out.append(os.path.realpath(os.path.normpath(os.path.join(self.outdir, fname))))
            self.fname_type.append(2)


    def EWS(self, V_hub_in):
        # Extreme wind shear: 6.3.2.6
        self.setup()

        T = 12
        t = np.linspace(0, T, num=(int(T/self.dt+1)))

        # constants from standard
        alpha = 0.2
        Beta = 6.4

        # Flow angle adjustments
        V_hub = V_hub_in*np.cos(self.Vert_Slope*np.pi/180)
        V_vert_mag = V_hub_in*np.sin(self.Vert_Slope*np.pi/180)

        sigma_1 = self.NTM(V_hub)

        # Contant variables
        V = np.zeros_like(t)+V_hub
        V_dir = np.zeros_like(t)
        V_vert = np.zeros_like(t)+V_vert_mag
        # shear_horz = np.zeros_like(t)
        shear_vert = np.zeros_like(t)+alpha
        V_gust = np.zeros_like(t)
        upflow = np.zeros_like(t)

        # Transient
        shear_lin_p = np.zeros_like(t)
        shear_lin_n = np.zeros_like(t)

        for i, ti in enumerate(t):
            shear_lin_p[i] = (2.5+0.2*Beta*sigma_1*(self.D/self.Sigma_1)**(1/4))*(1-np.cos(2*np.pi*ti/T))/V_hub
            shear_lin_n[i] = -1*(2.5+0.2*Beta*sigma_1*(self.D/self.Sigma_1)**(1/4))*(1-np.cos(2*np.pi*ti/T))/V_hub

        # Write Files
        self.fname_out = []
        self.fname_type = []
        if self.dir_change.lower() == 'both' or self.dir_change.lower() == '+':
            if self.shear_orient.lower() == 'both' or self.shear_orient.lower() == 'v':
                ## Vert
                fname = self.case_name + '_EWS_V_P_U%2.1f.wnd'%V_hub_in
                data = np.column_stack((t, V, V_dir, V_vert, np.zeros_like(t), shear_vert, shear_lin_p, V_gust, upflow))
                hd = []
                hd.append('! Exteme Vertical Wind Shear, positive\n')
                hd = self.heading_common(hd)
                hd = self.heading_variable(hd, V_hub_in=V_hub_in)
                self.write_wnd(fname, data, hd)
                self.fname_out.append(os.path.realpath(os.path.normpath(os.path.join(self.outdir, fname))))
                self.fname_type.append(2)

            if self.shear_orient.lower() == 'both' or self.shear_orient.lower() == 'h':
                # Horz
                fname = self.case_name + '_EWS_H_P_U%2.1f.wnd'%V_hub_in
                data = np.column_stack((t, V, V_dir, V_vert, shear_lin_p, shear_vert, np.zeros_like(t), V_gust, upflow))
                hd = []
                hd.append('! Exteme Horizontal Wind Shear, positive\n')
                hd = self.heading_common(hd)
                hd = self.heading_variable(hd, V_hub_in=V_hub_in)
                self.write_wnd(fname, data, hd)
                self.fname_out.append(os.path.realpath(os.path.normpath(os.path.join(self.outdir, fname))))
                self.fname_type.append(2)

        if self.dir_change.lower() == 'both' or self.dir_change.lower() == '-':
            if self.shear_orient.lower() == 'both' or self.shear_orient.lower() == 'v':
                ## Vert
                fname = self.case_name + '_EWS_V_N_U%2.1f.wnd'%V_hub_in
                data = np.column_stack((t, V, V_dir, V_vert, np.zeros_like(t), shear_vert, shear_lin_n, V_gust, upflow))
                hd = []
                hd.append('! Exteme Vertical Wind Shear, negative\n')
                hd = self.heading_common(hd)
                hd = self.heading_variable(hd, V_hub_in=V_hub_in)
                self.write_wnd(fname, data, hd)
                self.fname_out.append(os.path.realpath(os.path.normpath(os.path.join(self.outdir, fname))))
                self.fname_type.append(2)

            if self.shear_orient.lower() == 'both' or self.shear_orient.lower() == 'h':
                # Horz
                fname = self.case_name + '_EWS_H_N_U%2.1f.wnd'%V_hub_in
                data = np.column_stack((t, V, V_dir, V_vert, shear_lin_n, shear_vert, np.zeros_like(t), V_gust, upflow))
                hd = []
                hd.append('! Exteme Horizontal Wind Shear, negative\n')
                hd = self.heading_common(hd)
                hd = self.heading_variable(hd, V_hub_in=V_hub_in)
                self.write_wnd(fname, data, hd)
                self.fname_out.append(os.path.realpath(os.path.normpath(os.path.join(self.outdir, fname))))
                self.fname_type.append(2)

    def heading_common(self, hd):
        hd.append('! IEC Turbine Class %s, IEC Turbulence Category %s\n'%(self.Turbine_Class, self.Turbulence_Class))
        hd.append('! '+('%f'%(self.D)).ljust(14) + ' ' + 'Rotor_Diameter'.ljust(14) + ' - rotor diameter (m)\n')
        hd.append('! '+('%f'%(self.z_hub)).ljust(14) + ' ' + 'hub_height'.ljust(14) + ' - hub height of the wind turbine (m)\n')
        hd.append('! '+('%f'%(self.Sigma_1)).ljust(14) + ' ' + 'Sigma_1'.ljust(14) + ' - logitudinal turbulence scale parameter (m)\n')
        return hd
    
    def heading_variable(self, hd, V_hub_in=[], sigma_1=[]):
        if sigma_1:
            hd.append('! '+('%f'%(sigma_1)).ljust(14) + ' ' + 'sigma_1'.ljust(14) + ' - turbulence standard deviation\n')
        if V_hub_in:
            hd.append('! '+('%f'%(V_hub_in)).ljust(14) + ' ' + 'V_hub'.ljust(14) + ' - wind speed at hub height (m/s)\n')

        return hd

    def write_wnd(self, fname, data, hd):

        # Make sure directory exist
        if not os.path.isdir(self.outdir):
            os.makedirs(self.outdir)

        # Move transient event to user defined time
        data[:,0] += self.TStart
        data = np.vstack((data[0,:], data, data[-1,:]))
        data[0,0] = self.T0
        data[-1,0] = self.TF

        # Headers
        hd1 = ['Time', 'Wind', 'Wind', 'Vertical', 'Horiz.', 'Pwr. Law', 'Lin. Vert.', 'Gust', 'Upflow']
        hd2 = ['', 'Speed', 'Dir', 'Speed', 'Shear', 'Vert. Shr', 'Shear', 'Speed', 'Angle']
        hd3 = ['(s)', '(m/s)', '(deg)', '(m/s)', '(-)', '(-)', '(-)', '(m/s)', '(deg)']

        self.fpath = os.path.join(self.outdir, fname)
        fid = open(self.fpath, 'w')

        fid.write('! Wind file generated by pyIECWind - IEC 61400-1 3rd Edition\n')
        for ln in hd:
            fid.write(ln)
        fid.write('! ---------------------------------------------------------------\n')
        fid.write('! '+''.join([('%s' % val).center(12) for val in hd1]) + '\n')
        fid.write('! '+''.join([('%s' % val).center(12) for val in hd2]) + '\n')
        fid.write('! '+''.join([('%s' % val).center(12) for val in hd3]) + '\n')
        for row in data:
            fid.write('  '+''.join([('%.6f' % val).center(12) for val in row]) + '\n')

        fid.close()

    def execute(self, Vtype, V_hub):

        if 'EOG' in Vtype:
            self.EOG(V_hub)
        if 'EDC' in Vtype:
            self.EDC(V_hub)
        if 'ECD' in Vtype:
            self.ECD(V_hub)
        if 'EWS' in Vtype:
            self.EWS(V_hub)

        return self.fname_out, self.fname_type
        

class pyIECWind_turb():

    def __init__(self):

        # Defaults
        self.seed             = np.random.uniform(1, 1e8)
        self.Turbulence_Class = 'B'  # IEC Turbulance Class
        self.z_hub            = 90.  # wind turbine hub height (m)
        self.D                = 126. # rotor diameter (m)
        self.PLExp            = 0.2
        self.AnalysisTime     = 720.
        self.debug_level      = 0
        self.overwrite        = True

    def setup(self):
        turbsim_vt = turbsiminputs()
        turbsim_vt.runtime_options.RandSeed1  = self.seed
        turbsim_vt.runtime_options.WrADTWR    = False
        turbsim_vt.tmspecs.AnalysisTime       = self.AnalysisTime
        turbsim_vt.tmspecs.HubHt              = self.z_hub # wind grid centered at hub height
        turbsim_vt.tmspecs.GridHeight         = (self.z_hub - 1.) * 2.0 # wind grid stops 1 meter above the ground
        turbsim_vt.tmspecs.GridWidth          = (self.z_hub - 1.) * 2.0 # squared wind grid
        turbsim_vt.tmspecs.NumGrid_Z          = 25
        turbsim_vt.tmspecs.NumGrid_Y          = 25
        turbsim_vt.tmspecs.HFlowAng           = 0.0
        turbsim_vt.tmspecs.VFlowAng           = 0.0
        turbsim_vt.metboundconds.TurbModel    = '"IECKAI"'
        turbsim_vt.metboundconds.UserFile     = '"unused"'
        turbsim_vt.metboundconds.IECturbc     = self.Turbulence_Class
        turbsim_vt.metboundconds.IEC_WindType = self.IEC_WindType
        turbsim_vt.metboundconds.ETMc         = '"default"'
        turbsim_vt.metboundconds.WindProfileType = '"PL"'
        turbsim_vt.metboundconds.ProfileFile  = '"unused"'
        turbsim_vt.metboundconds.RefHt        = self.z_hub
        turbsim_vt.metboundconds.URef         = self.Uref
        turbsim_vt.metboundconds.PLExp        = self.PLExp
        
        turbsim_vt.noniecboundconds.Latitude  = '"default"'
        turbsim_vt.noniecboundconds.RICH_NO   = 0.05
        turbsim_vt.noniecboundconds.UStar     = '"default"'
        turbsim_vt.noniecboundconds.ZI        = '"default"'
        turbsim_vt.noniecboundconds.PC_UW     = '"default"'
        turbsim_vt.noniecboundconds.PC_UV     = '"default"'
        turbsim_vt.noniecboundconds.PC_VW     = '"default"'

        turbsim_vt.spatialcoherance.InCDec1   = '"default"'
        turbsim_vt.spatialcoherance.InCDec2   = '"default"'
        turbsim_vt.spatialcoherance.InCDec3   = '"default"'
        turbsim_vt.spatialcoherance.CohExp    = '"default"'
        
        
        
        return turbsim_vt

    def execute(self, IEC_WindType, Uref, ver='Turbsim'):
        self.IEC_WindType = IEC_WindType
        self.Uref = Uref

        turbsim_vt = self.setup()
        writer = TurbsimBuilder()
        if ver.lower() == 'turbsim':
            wrapper = Turbsim_wrapper()
        if ver.lower() =='pyturbsim':
            wrapper = pyTurbsim_wrapper()

        # if self.case_name[-3:] != '.in':
        #     self.case_name = self.case_name + '.in'
        # self.case_name += '_U%1.1f'%self.Uref + '_Seed%1.1f'%self.seed
        # self.case_name += '_U%d'%self.Uref + '_Seed%d.in'%self.seed

        case_name = self.case_name + '_' + IEC_WindType + '_U%1.6f'%self.Uref + '_Seed%1.1f'%self.seed
        
        tsim_input_file = case_name + '.in'
        wind_file_out   = case_name + '.bts'
        
        wind_file_out_abs = os.path.realpath(os.path.normpath(os.path.join(self.outdir, wind_file_out)))

        # If wind file already exists and overwriting is turned off, skip wind file write
        if os.path.exists(os.path.join(self.outdir, wind_file_out)) and not self.overwrite:
            return wind_file_out_abs, 3

        # Run wind file generation
        else:
            writer.turbsim_vt = turbsim_vt
            writer.run_dir = self.outdir
            writer.tsim_input_file = tsim_input_file
            writer.execute()

            wrapper.turbsim_input = os.path.realpath(os.path.join(writer.run_dir, writer.tsim_input_file))
            wrapper.run_dir = writer.run_dir
            wrapper.turbsim_exe = self.Turbsim_exe
            wrapper.debug_level = self.debug_level
            wrapper.execute()

            return wind_file_out_abs, 3


def example_ExtremeWind():

    iec = pyIECWind_extreme()
    iec.Turbine_Class = 'I'     # IEC Wind Turbine Class
    iec.Turbulence_Class = 'A'  # IEC Turbulance Class
    iec.dt = 0.05               # Transient wind time step (s)
    iec.dir_change = 'both'     # '+','-','both': sign for transient events in EDC, EWS
    iec.z_hub = 30.             # wind turbine hub height (m)
    iec.D = 42.                 # rotor diameter (m)

    iec.case_name = 'test'
    iec.outdir = 'temp'

    V_hub = 25
    iec.execute('EWS', V_hub)

def example_TurbulentWind():
    iec = pyIECWind_turb()
    
    iec.Turbulence_Class = 'A'  # IEC Turbulance Class
    iec.z_hub = 90.             # wind turbine hub height (m)
    iec.D = 126.                 # rotor diameter (m)
    iec.AnalysisTime = 30.

    iec.outdir = 'temp'
    iec.case_name = 'turbsim_testing'
    iec.Turbsim_exe = 'C:/Users/egaertne/WT_Codes/Turbsim_v2.00.07/bin/TurbSim_x64.exe'
    iec.debug_level = 1

    IEC_WindType = 'NTM'
    Uref = 10.

    iec.execute(IEC_WindType, Uref)


if __name__=="__main__":

    example_ExtremeWind()
    # example_TurbulentWind()



'''
A class for setup and execution of OpenFAST for testing and verification of the ROSCO 
controller. 


Run ROSCO and test against baseline results:
    - set up case(s) using CaseLibrary from aeroelasticse
    - run fast simulations
    - evaluate & compare results with pCrunch
    - report results to user


'''

import numpy as np
import os, platform
import glob
import multiprocessing as mp

import ROSCO_toolbox.ofTools.fast_io.read_fast_input as fast_io
from ROSCO_toolbox.ofTools.fast_io.FAST_reader import InputReader_OpenFAST
from ROSCO_toolbox.ofTools.case_gen.CaseGen_IEC import CaseGen_IEC
from ROSCO_toolbox.ofTools.case_gen.runFAST_pywrapper import runFAST_pywrapper_batch
from matplotlib.backends.backend_pdf import FigureCanvasPdf, PdfPages
from ROSCO_toolbox.ofTools.fast_io import output_processing
import matplotlib.pyplot as plt



class ROSCO_testing():
    '''
    Primary ROSCO test scripts
    '''

    def __init__(self, **kwargs):


        # Setup simulation parameters
        self.runDir = os.path.join(os.path.dirname( os.path.realpath(__file__) ), 'testing' )   # directory to run simulations in
        self.wind_dir = None
        self.namebase = 'ROtest'    # root name for output simulations
        self.FAST_exe = 'openfast_single'       # name of openfast executable (may need full path)
        self.Turbsim_exe = 'turbsim_single'     # name of turbsim executable
        self.FAST_ver = 'OpenFAST'  # Fast version
        # Path to ROSCO controller - default to ROSCO Toolbox submodule
        try:
            self.rosco_path = glob.glob(os.path.join(os.path.dirname(os.path.realpath(__file__)),'../ROSCO/build/libdiscon.*'))[0]
        except:
            print('No compiled ROSCO version found, please provide ROSCO_testing.rosco_path.')
        self.dev_branch = True      # openfast dev branch?
        self.debug_level = 2        # debug level. 0 - no outputs, 1 - minimal outputs, 2 - all outputs
        self.overwrite = False      # overwrite existing files? 
        self.cores = 4              # number of cores to use
        self.mpi_run = False
        self.mpi_comm_map_down = []
        self.outfile_fmt = 2 # 1 = .txt, 2 = binary, 3 = both
        self.comp_dir = None

        # Setup turbine parameters 
        #  - Default to NREL 5MW 
        self.Turbine_Class = 'I'
        self.Turbulence_Class = 'A'
        self.FAST_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../Test_Cases/NREL-5MW')
        self.FAST_InputFile = 'NREL-5MW.fst'

        # Desired output channesl
        self.var_out = [
                        'BldPitch1', 'BldPitch2', 'BldPitch3', 'GenTq', 'GenPwr', 'RotSpeed', 'GenSpeed',
                        'TipDxc1', 'TipDyc1', 'TipDzc1', 'TipDxc2',
                        'TipDyc2', 'TipDzc2', 'TipDxc3', 'TipDyc3', 'TipDzc3',
                        'RootMxc1', 'RootMyc1', 'RootMzc1',
                        'RootMxc2', 'RootMyc2', 'RootMzc2',
                        'RootMxc3', 'RootMyc3', 'RootMzc3',
                        'RootMxb1', 'RootMyb1', 'RootMzb1',
                        'RootMxb2', 'RootMyb2', 'RootMzb2',
                        'RootMxb3', 'RootMyb3', 'RootMzb3',
                        'TwrBsMxt', 'TwrBsMyt', 'TwrBsMzt',
                        'TwrBsFxt', 'TwrBsFyt', 'TwrBsFzt',
                        'Wind1VelX', 'Wind1VelY', 'Wind1VelZ',
                        'RtVAvgxh', 'RtVAvgyh', 'RtVAvgzh'
                        ]
                        
        if self.cores > mp.cpu_count():
            self.parallel_cores = mp.cpu_count()

        for (k, w) in kwargs.items():
            try:
                setattr(self, k, w)
            except:
                pass

        super(ROSCO_testing, self).__init__()

    def ROSCO_Test_lite(self, more_case_inputs={}, U=[]):
        '''
        DLC 1.1 - 5 wind speeds, 60s

        Parameters:
        -----------
        more_case_inputs: dict
            Additional case inputs
        U: list
            List of wind inputs
        '''

        # Check for time and wind inputs
        if ('Fst','TMax') in more_case_inputs.keys():
            self.TMax = np.max(more_case_inputs[('Fst','TMax')]['vals'])
        else:
            self.TMax = 330
        
        if len(U) > 0:
            WindSpeeds = U
        else:
            WindSpeeds = [5, 8, 11, 14, 17]

        fastRead = InputReader_OpenFAST(
            FAST_ver=self.FAST_ver, dev_branch=self.dev_branch)
        fastRead.FAST_InputFile =  self.FAST_InputFile   # FAST input file (ext=.fst)
        # Path to fst directory files
        fastRead.FAST_directory = self.FAST_directory

        # Read FAST inputs for generating cases
        fastRead.execute()

        # Start near the steady state, controller should be able to handle startup transients.
        iec = CaseGen_IEC()
        iec.init_cond[("ElastoDyn", "RotSpeed")] = {'U':  [2, 30]}
        iec.init_cond[("ElastoDyn", "RotSpeed")]['val'] = np.ones(
            [2]) * fastRead.fst_vt['ElastoDyn']['RotSpeed'] * .75
        iec.init_cond[("ElastoDyn", "BlPitch1")] = {'U':  [2, 30]}
        iec.init_cond[("ElastoDyn", "BlPitch1")]['val'] = np.ones([2]) * 0
        iec.init_cond[("ElastoDyn", "BlPitch2")] = iec.init_cond[("ElastoDyn", "BlPitch1")]
        iec.init_cond[("ElastoDyn", "BlPitch3")] = iec.init_cond[("ElastoDyn", "BlPitch1")]
        iec.Turbine_Class = self.Turbine_Class
        iec.Turbulence_Class = self.Turbulence_Class
        iec.D = fastRead.fst_vt['ElastoDyn']['TipRad']*2.
        iec.z_hub = fastRead.fst_vt['InflowWind']['RefHt']
        iec.TMax = self.TMax

        iec.dlc_inputs = {}
        iec.dlc_inputs['DLC'] = [1.1]  # ,6.1,6.3]
        iec.dlc_inputs['U'] = [WindSpeeds]
        iec.dlc_inputs['Seeds'] = [[971231]]
        iec.dlc_inputs['Yaw'] = [[]]
        iec.transient_dir_change = '-'  # '+','-','both': sign for transient events in EDC, EWS
        iec.transient_shear_orientation = 'v'  # 'v','h','both': vertical or horizontal shear for EWS

        if self.wind_dir:
            iec.wind_dir = self.wind_dir
        else:
            iec.wind_dir = os.path.join(self.runDir, 'wind')
            
        iec.case_name_base = self.namebase
        iec.Turbsim_exe = self.Turbsim_exe
        iec.debug_level = self.debug_level
        iec.cores = self.cores
        iec.run_dir = self.runDir
        iec.overwrite = self.overwrite
        # iec.overwrite       = False
        if self.cores > 1:
            iec.parallel_windfile_gen = True
        else:
            iec.parallel_windfile_gen = False

        # mpi_run = False
        if self.mpi_run:
            iec.mpi_run = mpi_run
            iec.comm_map_down = mpi_comm_map_down

        case_inputs = {}
        case_inputs[("Fst", "TMax")] = {'vals': [self.TMax], 'group': 0}
        case_inputs[("Fst", "OutFileFmt")] = {'vals': [self.outfile_fmt], 'group': 0}

        case_inputs[('ServoDyn', 'GenTiStr')] = {'vals': ['True'], 'group': 0}
        case_inputs[('ServoDyn', 'GenTiStp')] = {'vals': ['True'], 'group': 0}
        case_inputs[('ServoDyn', 'SpdGenOn')] = {'vals': [0.], 'group': 0}
        case_inputs[('ServoDyn', 'TimGenOn')] = {'vals': [0.], 'group': 0}
        case_inputs[('ServoDyn', 'DLL_FileName')] = {'vals': [self.rosco_path], 'group': 0}
        case_inputs[('ServoDyn', 'DLL_DT')] = {'vals': ['"default"'], 'group': 0}

        case_inputs[("AeroDyn15", "WakeMod")] = {'vals': [1], 'group': 0}
        case_inputs[("AeroDyn15", "AFAeroMod")] = {'vals': [2], 'group': 0}
        case_inputs[("AeroDyn15", "TwrPotent")] = {'vals': [0], 'group': 0}
        case_inputs[("AeroDyn15", "TwrShadow")] = {'vals': ['False'], 'group': 0}
        case_inputs[("AeroDyn15", "TwrAero")] = {'vals': ['False'], 'group': 0}
        case_inputs[("AeroDyn15", "SkewMod")] = {'vals': [1], 'group': 0}
        case_inputs[("AeroDyn15", "TipLoss")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "HubLoss")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "TanInd")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "AIDrag")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "TIDrag")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "IndToler")] = {'vals': [1.e-5], 'group': 0}
        case_inputs[("AeroDyn15", "MaxIter")] = {'vals': [5000], 'group': 0}
        case_inputs[("AeroDyn15", "UseBlCm")] = {'vals': ['True'], 'group': 0}

        if more_case_inputs:
            case_inputs.update(more_case_inputs)
        
        # generate cases
        case_list, case_name_list, _ = iec.execute(case_inputs=case_inputs)

        # Ensure proper output channels
        var_out = self.var_out

        channels = {}
        for var in var_out:
            channels[var] = True

        # Set up FAST Sims
        fastBatch = runFAST_pywrapper_batch()
        fastBatch.FAST_ver = self.FAST_ver
        fastBatch.FAST_exe = self.FAST_exe   # Path to executable
        fastBatch.FAST_runDirectory = self.runDir
        fastBatch.FAST_InputFile = self.FAST_InputFile  # FAST input file (ext=.fst)
        fastBatch.FAST_directory = self.FAST_directory   # Path to fst directory files
        fastBatch.debug_level = self.debug_level
        fastBatch.dev_branch = self.dev_branch

        fastBatch.case_list = case_list
        fastBatch.case_name_list = case_name_list
        fastBatch.channels = channels

        # Check if simulation has been run
        if self.outfile_fmt == 1:
            outfile_ext = '.out'
        elif self.outfile_fmt == 2:
            outfile_ext = '.outb'
        elif self.outfile_fmt == 3:
            outfile_ext = '.outb'
        else:
            print('Warning, outfile format may be invalid. Attempting to read binary outputs.')
            outfile_ext = 'outb'
        outFileNames = [os.path.join(fastBatch.FAST_runDirectory, case_name + outfile_ext)
                        for case_name in case_name_list]
        outFileThere = [os.path.exists(outFileName) for outFileName in outFileNames]

        # Run simulations if they're not all there or if you want to overwrite
        if not all(outFileThere) or self.overwrite:
            if self.cores > 1:
                fastBatch.run_multi(self.cores)
            else:
                fastBatch.run_serial()

        self.print_results(outFileNames)


    def ROSCO_Test_heavy(self, more_case_inputs={}, U=[]):
        '''
        Run extensive DLCs for ROSCO
            - DLC 1.3 - Cutin-Cutout, 2 seeds
            - DLC 1.4 - 2 wind speeds

        more_case_inputs: dict
            Additional case inputs
        U: list-like
            List like with two lists of wind speeds, first entry for DLC 1.3 and second entry for DLC 1.4
        '''

        # Check for time and wind inputs
        if ('Fst','TMax') in more_case_inputs.keys():
            self.TMax = np.max(more_case_inputs[('Fst','TMax')]['vals'])
        else:
            self.TMax = 630
        
        if len(U) > 0:
            WindSpeeds = U
            if len(U) != 2:
                ValueError('For a user defined input, U, two sets of wind speeds must be defined.')
        else:
            WindSpeeds = [[4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24], [8.88, 12.88]]

        fastRead = InputReader_OpenFAST(
            FAST_ver=self.FAST_ver, dev_branch=self.dev_branch)
        fastRead.FAST_InputFile = self.FAST_InputFile   # FAST input file (ext=.fst)
        # Path to fst directory files
        fastRead.FAST_directory = self.FAST_directory

        # Read FAST inputs for generating cases
        fastRead.execute()

        # Start near the steady state, controller should be able to handle startup transients.
        iec = CaseGen_IEC()
        iec.init_cond[("ElastoDyn", "RotSpeed")] = {'U':  [2, 30]}
        iec.init_cond[("ElastoDyn", "RotSpeed")]['val'] = np.ones(
            [2]) * fastRead.fst_vt['ElastoDyn']['RotSpeed'] * .75
        iec.init_cond[("ElastoDyn", "BlPitch1")] = {'U':  [2, 30]}
        iec.init_cond[("ElastoDyn", "BlPitch1")]['val'] = np.ones([2]) * 0
        iec.init_cond[("ElastoDyn", "BlPitch2")] = iec.init_cond[("ElastoDyn", "BlPitch1")]
        iec.init_cond[("ElastoDyn", "BlPitch3")] = iec.init_cond[("ElastoDyn", "BlPitch1")]

        # waves
        if self.FAST_InputFile == 'IEA-15-240-RWT-UMaineSemi.fst':
            iec.init_cond[('HydroDyn','WaveHs')] = {'U': [4., 6., 8., 10., 12., 14., 16., 18., 20., 22., 24.]}
            iec.init_cond[('HydroDyn','WaveHs')]['val'] = [1.102,1.179,1.316,1.537,1.836,2.188,2.598,3.061,3.617,4.027,4.516]
            iec.init_cond[('HydroDyn','WaveTp')] = {'U': [4., 6., 8., 10., 12., 14., 16., 18., 20., 22., 24.]}
            iec.init_cond[('HydroDyn','WaveTp')]['val'] = [8.515,8.310,8.006,7.651,7.441,7.461,7.643,8.047,8.521,8.987,9.452]

        iec.Turbine_Class = self.Turbine_Class
        iec.Turbulence_Class = self.Turbulence_Class
        iec.D = fastRead.fst_vt['ElastoDyn']['TipRad']*2.
        iec.z_hub = fastRead.fst_vt['InflowWind']['RefHt']
        iec.TMax = self.TMax

        iec.dlc_inputs = {}
        iec.dlc_inputs['DLC'] = [1.3, 1.4] 
        iec.dlc_inputs['U'] = WindSpeeds
        iec.dlc_inputs['Seeds'] = [[991235, 5123], []]
        iec.dlc_inputs['Yaw'] = [[], []]
        iec.transient_dir_change = '-'  # '+','-','both': sign for transient events in EDC, EWS
        iec.transient_shear_orientation = 'v'  # 'v','h','both': vertical or horizontal shear for EWS
        iec.uniqueSeeds = True
        iec.uniqueWaveSeeds = True

        if self.wind_dir:
            iec.wind_dir = self.wind_dir
        else:
            iec.wind_dir = os.path.join(self.runDir, 'wind')
        iec.case_name_base = self.namebase
        iec.Turbsim_exe = self.Turbsim_exe
        iec.debug_level = self.debug_level
        iec.cores = self.cores
        iec.run_dir = os.path.join(self.runDir)
        iec.overwrite = self.overwrite
        # iec.overwrite       = False
        if self.cores > 1:
            iec.parallel_windfile_gen = True
        else:
            iec.parallel_windfile_gen = False

        # mpi_run = False
        if self.mpi_run:
            iec.mpi_run = mpi_run
            iec.comm_map_down = mpi_comm_map_down

        case_inputs = {}
        case_inputs[("Fst", "TMax")] = {'vals': [self.TMax], 'group': 0}
        case_inputs[("Fst", "OutFileFmt")] = {'vals': [self.outfile_fmt], 'group': 0}

        case_inputs[('ServoDyn', 'GenTiStr')] = {'vals': ['False'], 'group': 0}
        case_inputs[('ServoDyn', 'GenTiStp')] = {'vals': ['True'], 'group': 0}
        case_inputs[('ServoDyn', 'SpdGenOn')] = {'vals': [0.], 'group': 0}
        case_inputs[('ServoDyn', 'DLL_FileName')] = {'vals': [self.rosco_path], 'group': 0}

        case_inputs[("AeroDyn15", "WakeMod")] = {'vals': [1], 'group': 0}
        case_inputs[("AeroDyn15", "AFAeroMod")] = {'vals': [2], 'group': 0}
        case_inputs[("AeroDyn15", "TwrPotent")] = {'vals': [0], 'group': 0}
        case_inputs[("AeroDyn15", "TwrShadow")] = {'vals': ['False'], 'group': 0}
        case_inputs[("AeroDyn15", "TwrAero")] = {'vals': ['False'], 'group': 0}
        case_inputs[("AeroDyn15", "SkewMod")] = {'vals': [1], 'group': 0}
        case_inputs[("AeroDyn15", "TipLoss")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "HubLoss")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "TanInd")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "AIDrag")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "TIDrag")] = {'vals': ['True'], 'group': 0}
        case_inputs[("AeroDyn15", "IndToler")] = {'vals': [1.e-5], 'group': 0}
        case_inputs[("AeroDyn15", "MaxIter")] = {'vals': [5000], 'group': 0}
        case_inputs[("AeroDyn15", "UseBlCm")] = {'vals': ['True'], 'group': 0}

        if self.FAST_InputFile == 'IEA-15-240-RWT-UMaineSemi.fst':
            case_inputs[("HydroDyn", "WaveMod")] = {'vals': [2], 'group': 0}

        if more_case_inputs:
            case_inputs.update(more_case_inputs)

        case_list, case_name_list, _ = iec.execute(case_inputs=case_inputs)
        
        # Ensure proper output channels
        var_out = self.var_out

        channels = {}
        for var in var_out:
            channels[var] = True

        # Set up FAST Sims, move setup up
        fastBatch = runFAST_pywrapper_batch()
        fastBatch.FAST_ver = self.FAST_ver
        fastBatch.FAST_exe = self.FAST_exe   # Path to executable
        fastBatch.FAST_runDirectory = self.runDir
        fastBatch.FAST_InputFile = self.FAST_InputFile  # FAST input file (ext=.fst)
        fastBatch.FAST_directory = self.FAST_directory   # Path to fst directory files
        fastBatch.debug_level = self.debug_level
        fastBatch.dev_branch = self.dev_branch

        fastBatch.case_list = case_list
        fastBatch.case_name_list = case_name_list
        fastBatch.channels = channels

        # Check if simulation has been run
        if self.outfile_fmt == 1:
            outfile_ext = '.out'
        elif self.outfile_fmt == 2:
            outfile_ext = '.outb'
        elif self.outfile_fmt == 3:
            outfile_ext = '.outb'
        else:
            print('Warning, outfile format may be invalid. Attempting to read binary outputs.')
            outfile_ext = 'outb'
        outFileNames = [os.path.join(fastBatch.FAST_runDirectory, case_name + outfile_ext)
                        for case_name in case_name_list]
        outFileThere = [os.path.exists(outFileName) for outFileName in outFileNames]

        # Run simulations if they're not all there or if you want to overwrite
        if not all(outFileThere) or self.overwrite:
            if self.cores > 1:
                fastBatch.run_multi(self.cores)
            else:
                fastBatch.run_serial()

        self.print_results(outFileNames)

    def ROSCO_Controller_Comp(self, controller_paths, testtype='light', more_case_inputs={}, U=[]):
        '''
        Heavy or light testing for n controllers, n = len(controller_paths)

        Parameters:
        ----------
        controller_paths: list
            list of paths to .dlls
        testtype: str
            type of test to run. 'light' or 'heavy
        '''
        # Save initial run directory
        run_dir_init = self.runDir
        wind_dir_init = self.wind_dir
        for ci, path in enumerate(controller_paths):
            # specify rosco path
            self.rosco_path = path
            # temporarily change run directories
            self.runDir = os.path.join(run_dir_init,'controller_{}'.format(ci)) # specific directory for each controller
            self.wind_dir = os.path.join(run_dir_init, 'wind')  # wind in base runDir

            if testtype.lower() == 'light':
                self.ROSCO_Test_lite(more_case_inputs=more_case_inputs, U=U)
            elif testtype.lower() == 'heavy':
                self.ROSCO_Test_heavy()
            else:
                raise ValueError('{} is an invalid testtype for controller comparison'.format(testtype))

        # reset self
        self.runDir = run_dir_init
        self.wind_dir = wind_dir_init
    
    def ROSCO_DISCON_Comp(self, DISCON_filenames, testtype='light', more_case_inputs={}, U=[]):
        '''
        Heavy or light testing for n DISCON.IN files, n = len(DISCON_paths)

        Parameters:
        ----------
        controller_paths: list
            list of paths to .dlls
        testtype: str
            type of test to run. 'light' or 'heavy
        '''

        # Save initial run directory
        run_dir_init = self.runDir
        wind_dir_init = self.wind_dir
        for ci, discon in enumerate(DISCON_filenames):
            # temporarily change run directories
            self.runDir = os.path.join(run_dir_init, 'controller_{}'.format(ci))
            self.wind_dir = os.path.join(run_dir_init, 'wind')  # wind in base runDir

            # Point to different DISCON.IN files using more_case_inputs
            # Control (DISCON) Inputs
            discon_vt = ROSCO_utilities.read_DISCON(discon)
            for discon_input in discon_vt:
                more_case_inputs[('DISCON_in',discon_input)] = {'vals': [discon_vt[discon_input]], 'group': 0}

            self.wind_dir = os.path.join(run_dir_init, 'wind')  # wind in base runDir

            if testtype.lower() == 'light':
                self.ROSCO_Test_lite(more_case_inputs=more_case_inputs, U=U)
            elif testtype.lower() == 'heavy':
                self.ROSCO_Test_heavy(more_case_inputs=more_case_inputs, U=U)
            else:
                raise ValueError('{} is an invalid testtype for DISCON comparison'.format(testtype))

        # reset self
        self.runDir = run_dir_init
        self.wind_dir = wind_dir_init

    def print_results(self,outfiles):

        figs_fname = 'test_outputs.pdf'

        # remove initial transients if more than two minutes if simulation time
        if self.TMax > 120:
            tmin = 60
        else:
            tmin = 0

        # Comparison plot
        if self.comp_dir:
            tmin = 100      # if comparing, I'd like to start the comparison at 100 seconds
            op_comp = output_processing.output_processing()
            outfiles_comp = [os.path.join(self.comp_dir,os.path.split(of)[1]) for of in outfiles]
            try:
                FAST_Comp = op_comp.load_fast_out(outfiles_comp, tmin=tmin)
                figs_fname = 'comp_outputs.pdf'
            except:
                print('Could not load openfast outputs for comparison, only plotting current test')
                tmin = 0
                figs_fname = 'test_outputs.pdf'


        op = output_processing.output_processing()
        FAST_Output = op.load_fast_out(outfiles, tmin=tmin)

                
        with PdfPages(os.path.join(self.runDir,figs_fname)) as pdf:
            for i_out, fast_out in enumerate(FAST_Output):
                if self.FAST_InputFile == 'NREL-5MW.fst':
                    plots2make = {'Baseline': ['Wind1VelX', 'GenPwr', 'RotSpeed', 'BldPitch1', 'GenTq']}
                else:
                    plots2make = {'Baseline': ['Wind1VelX', 'GenPwr', 'RotSpeed', 'BldPitch1', 'GenTq','PtfmPitch','TwrBsMyt']}

                numplots    = len(plots2make)
                maxchannels = np.max([len(plots2make[key]) for key in plots2make.keys()])
                fig = plt.figure(figsize=(8,6), constrained_layout=True)
                gs_all = fig.add_gridspec(1, numplots)
                for pnum, (gs, pname) in enumerate(zip(gs_all, plots2make.keys())):
                    gs0 = gs.subgridspec(len(plots2make[pname]),1)
                    for cid, channel in enumerate(plots2make[pname]):
                        subplt = fig.add_subplot(gs0[cid])
                        try:
                            subplt.plot(fast_out['Time'], fast_out[channel])
                            if self.comp_dir:
                                subplt.plot(FAST_Comp[i_out]['Time'], FAST_Comp[i_out][channel])

                            unit_idx = fast_out['meta']['channels'].index(channel)
                            subplt.set_ylabel('{:^} \n ({:^})'.format(
                                                channel,
                                                fast_out['meta']['attribute_units'][unit_idx]))
                            subplt.grid(True)
                            subplt.set_xlabel('Time (s)')
                        except:
                            print('Cannot plot {}'.format(channel))
                        if cid == 0:
                            subplt.set_title(pname)
                        if cid != len(plots2make[pname])-1:
                            subplt.axes.get_xaxis().set_visible(False)

                    plt.suptitle(fast_out['meta']['name'])
                pdf.savefig(fig)
                plt.close()


if __name__=='__main__':
    rt = ROSCO_testing()


    ## =================== INITIALIZATION ===================
    # Setup simulation parameters
    rt.namebase = 'IEA-15MW'     # Base name for FAST files 
    rt.FAST_exe = 'openfast'     # OpenFAST executable path
    rt.Turbsim_exe = 'turbsim'   # Turbsim executable path
    rt.FAST_ver = 'OpenFAST'     # FAST version
    # path to compiled ROSCO controller
    if platform.system() == 'Windows':
        rt.rosco_path = os.path.join(os.getcwd(), '../ROSCO/build/libdiscon.dll')
    elif platform.system() == 'Darwin':
        rt.rosco_path = os.path.join(os.getcwd(), '../ROSCO/build/libdiscon.dylib')
    else:
        rt.rosco_path = os.path.join(os.getcwd(), '../ROSCO/build/libdiscon.so')
    rt.dev_branch = True         # dev branch of Openfast?
    rt.debug_level = 2           # debug level. 0 - no outputs, 1 - minimal outputs, 2 - all outputs
    rt.overwrite = True          # overwite fast sims?
    rt.cores = 4                 # number of cores if multiprocessings
    rt.mpi_run = False           # run using mpi
    rt.mpi_comm_map_down = []    # core mapping for MPI
    rt.outfile_fmt = 2           # 1 = .txt, 2 = binary, 3 = both
    rt.dev_branch= 'True'

    # Setup turbine
    rt.Turbine_Class = 'I'
    rt.Turbulence_Class = 'B'
    rt.FAST_directory = os.path.join(os.getcwd(), '../Test_Cases/IEA-15-240-RWT-UMaineSemi')
    rt.FAST_InputFile = 'IEA-15-240-RWT-UMaineSemi.fst'

    # Additional inputs 
    # ---- DT for this test! ----
    case_inputs={}
    case_inputs[('Fst', 'TMax')] = {'vals': [60], 'group': 0}
    case_inputs[('Fst', 'DT')] = {'vals': [0.01], 'group': 0}
    case_inputs[('Fst', 'CompElast')] = {'vals': [1], 'group': 0}

    case_inputs[('DISCON_in', 'PC_ControlMode')] = {'vals': [1], 'group': 0}
    case_inputs[('DISCON_in', 'PS_Mode')] = {'vals': [0], 'group': 0}
    case_inputs[('DISCON_in', 'VS_ControlMode')] = {'vals': [2], 'group': 0}
    case_inputs[('DISCON_in', 'WE_Mode')] = {'vals': [2], 'group': 0}

    # Wind Speeds
    U = [5, 9, 12, 15]

    # Run test
    rt.ROSCO_Test_lite(more_case_inputs=case_inputs, U=U)

